import os
import json
import time
import numpy as np
import pandas as pd
import seaborn as sb
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle

import constants, paths, ras_postprocessing, utils

import locale
locale.setlocale(locale.LC_ALL, '')
 
from paths import * 


def depthdamage(fn, output_dir):
	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(figsize=(7.5,4.5))
	
	### Read depth-damage functions CSV to dataframe
	dd_df = pd.read_csv(paths.depthdamage_funcs_uri)
	dd_df = dd_df[dd_df['DEPTH_FT']>=0]
	dd_df['DEPTH_M'] = dd_df['DEPTH_FT'] * constants.ft2meters

	### Plot curves
	axes.plot(dd_df['DEPTH_M'], dd_df['R12B'], 'ro-', label='Single-family home')
	axes.plot(dd_df['DEPTH_M'], dd_df['R21N'], 'bo-', label='Mobile home')
	axes.plot(dd_df['DEPTH_M'], dd_df['C1LB'], 'ko-', label='Commercial')

	### Set axis limits
	axes.set_xlim(-0.25, 5.35)
	axes.set_ylim(-5, 105)

	### Set axis labels
	axes.set_xlabel('Flood Depth (meters)')
	axes.set_ylabel('Property Damage (%)')

	### Add legend
	legend = axes.legend(loc='lower right', frameon=True)
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')

	### Save figure
	file = fn + '.pdf'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight')
	
	### Open figure
	time.sleep(0.5)
	os.system('open ' + uri)


def equityweights_v1(prop_vals, costs, weights, fn, output_dir):
	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(1, len(weights), figsize=(7.5,2.5))

	### Loop through weights
	for i, w in enumerate(weights):

		### Calculate equity-weighted utility
		lowvalprop_damagecost_ew = utils.ew_utility(prop_vals[0], costs, w)
		midvalprop_damagecost_ew = utils.ew_utility(prop_vals[1], costs, w)
		highvalprop_damagecost_ew = utils.ew_utility(prop_vals[2], costs, w)

		ymin = np.min([lowvalprop_damagecost_ew, 
					   midvalprop_damagecost_ew, 
					   highvalprop_damagecost_ew])
		ymax = np.max([lowvalprop_damagecost_ew, 
					   midvalprop_damagecost_ew, 
					   highvalprop_damagecost_ew])

		lowvalprop_damagecost_ew = (lowvalprop_damagecost_ew-ymin) / (ymax-ymin)
		midvalprop_damagecost_ew = (midvalprop_damagecost_ew-ymin) / (ymax-ymin)
		highvalprop_damagecost_ew = (highvalprop_damagecost_ew-ymin) / (ymax-ymin)

		### Plot curves
		axes[i].plot(costs, highvalprop_damagecost_ew, '-', 
			label=locale.currency(prop_vals[2], grouping=True)[:-3])
		axes[i].plot(costs, midvalprop_damagecost_ew, '-', 
			label=locale.currency(prop_vals[1], grouping=True)[:-3])
		axes[i].plot(costs, lowvalprop_damagecost_ew, '-', 
			label=locale.currency(prop_vals[0], grouping=True)[:-3])

		### Set axis limits
		axes[i].set_xlim(-500, 10**4 + 500)
		# axes[i].set_ylim(0, )

		### Set axis scale
		axes[i].set_yscale('log')

		### Set axis title
		axes[i].text(0.05, 0.95, r'$\eta$ = ' + str(w), 
			transform = axes[i].transAxes, va='top', ha='left')

		### Turn off minor ticks
		axes[i].tick_params(axis='y', which='minor', left='off')
		if i != 0:
			axes[i].tick_params(axis='y', which='major', left='off')

		### Set y-axis tick labels
		labels = []
		axes[i].set_yticklabels(labels)

		### Format x-axis tick labels 
		axes[i].set_xticks([0, 5000, 10000])
		fmt = '${x:,.0f}'
		tick = ticker.StrMethodFormatter(fmt)
		axes[i].xaxis.set_major_formatter(tick) 

	### Set y-axis tick labels
	axes[0].set_yticks(np.logspace(-3, 0, 6))
	labels = ['0.0', '0.2', '0.4', '0.6', '0.8', '1.0']
	axes[0].set_yticklabels(labels)

	### Set axis labels
	axes[1].set_xlabel('Avoided damage cost', labelpad=5)
	axes[0].set_ylabel('Equity-weighted utility\nof avoided damage cost', labelpad=10)

	### Add legend
	handles, labels = axes[0].get_legend_handles_labels()
	legend = axes[1].legend(handles[::-1], labels[::-1], loc='lower right', 
		frameon=True)	
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')

	### Save figure
	file = fn + '.pdf'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight')
	
	### Open figure
	time.sleep(0.5)
	os.system('open ' + uri)


def equityweights_v2(prop_vals, costs, weights, fn, output_dir):
	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots()

	### Loop through property values
	for i, prop_val in enumerate(prop_vals):
		u = []
		for n in weights:
			u.append(utils.ew_utility(prop_val, costs, n))
		prop_val = locale.currency(prop_val, grouping=True)[:-3]
		axes.plot(weights, u, '-', label='%s property' %prop_val)
		
	### Set axis limits
	axes.set_xlim(np.min(weights) - 0.1, np.max(weights) + 0.1)
	# axes.set_ylim(0, )

	### Set axis scale
	axes.set_yscale('log')

	### Set axis labels
	axes.set_xlabel(r'Equity weight ($\eta$)')
	axes.set_ylabel('Equity-weighted utility (u)')
	
	### Add legend
	legend = axes.legend(loc='upper right', frameon=True)	
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')

	### Save figure
	file = fn + '.pdf'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight')
	
	### Open figure
	time.sleep(0.5)
	os.system('open ' + uri)

	return None


def fitness_plot(df, fn, output_dir):

	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(2, 2, figsize=(7.5,6.5))
	plt.subplots_adjust(wspace=0.5)


	### Remove first line of df
	df = df[1:]

	### Set y-axis tick label format
	fmt = '${x:,.0f}'
	tick_fmt = ticker.StrMethodFormatter(fmt)
	axes[0,0].yaxis.set_major_formatter(tick_fmt)
	axes[0,1].yaxis.set_major_formatter(tick_fmt)
	axes[1,0].yaxis.set_major_formatter(tick_fmt)
	axes[1,1].yaxis.set_major_formatter(tick_fmt) 

	### Upper-left plot: Y = Fitness 1, X = Gens
	axes[0,0].set_ylabel('Avoided Damage Cost')
	
	x = df['GENERATION'].unique()
	y_max  = df.groupby(['GENERATION'])['FITNESS0'].max()
	y_mean = df.groupby(['GENERATION'])['FITNESS0'].mean()
	y_std  = df.groupby(['GENERATION'])['FITNESS0'].std()

	axes[0,0].plot(x, y_max)
	axes[0,0].plot(x, y_mean)
	axes[0,0].plot(x, y_std)

	### Upper-right plot: Y = Fitness 1, X = Fitness evals
	x = df.index
	y_max = df['FITNESS0'].expanding().max()
	y_mean = df['FITNESS0'].expanding().mean()
	y_std = df['FITNESS0'].expanding().std()

	axes[0,1].plot(x, y_max, label='Min/Max')
	axes[0,1].plot(x, y_mean, label='Mean')
	axes[0,1].plot(x, y_std, label='Std')

	### Lower-left plot: Y = Fitness 2, X = Gens
	axes[1,0].set_xlabel('Generation')
	axes[1,0].set_ylabel('Restoration Cost')

	x = df['GENERATION'].unique()	
	y_min  = df.groupby(['GENERATION'])['FITNESS3'].min()
	y_mean = df.groupby(['GENERATION'])['FITNESS3'].mean() 
	y_std  = df.groupby(['GENERATION'])['FITNESS3'].std()

	axes[1,0].plot(x, y_min)
	axes[1,0].plot(x, y_mean)
	axes[1,0].plot(x, y_std)

	### Lower-right plot: Y = Fitness 2, X = Fitness evals
	axes[1,1].set_xlabel('Fitness Evaluation')

	x = df.index
	y_min = df['FITNESS3'].expanding().min()
	y_mean = df['FITNESS3'].expanding().mean()
	y_std = df['FITNESS3'].expanding().std()

	axes[1,1].plot(x, y_min)
	axes[1,1].plot(x, y_mean)
	axes[1,1].plot(x, y_std)

	### Add legend
	legend = axes[0,1].legend(loc='upper right', frameon=True, 
							  bbox_to_anchor=(1.5,1))	
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')

	### Save figure
	file = fn + '.pdf'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight')
	
	### Open figure
	time.sleep(0.5)
	os.system('open ' + uri)

	return None


def diversity_plot(df, fn, output_dir):

	def indv_str2list(indv):
		return map(int, indv[1:-1].split(','))


	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots()

	### Calculate diversity
	x = df['GENERATION'].unique()
	y = pd.Series([])
	for g in x:
		indvs = df['INDIVIDUAL'][df['GENERATION']==g]

		gen_diversity = 0
		for indv0 in indvs:
			indv0 = indv_str2list(indv0)

			for indv1 in indvs:
				indv1 = indv_str2list(indv1)

				score = 0
				for attr0, attr1 in zip(indv0, indv1):
					if attr0 != attr1:
						score += 1

				score /= float(len(indv0))
				gen_diversity += score

		gen_diversity /= len(indvs)**2 * 2
		y = y.append(pd.Series([gen_diversity], index=[len(y)]))


	### Plot data
	axes.plot(x, y, 'k-o')

	### Axis labels
	axes.set_xlabel('Generation')
	axes.set_ylabel('Genetic Diversity')

	### Set y-axis limits
	axes.set_ylim(-0.05,0.55)

	### Save figure
	file = fn + '.pdf'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight')
	
	### Open figure
	time.sleep(0.5)
	os.system('open ' + uri)

	return None


def violin_plot(df, fn, output_dir):
	
	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(1, 2, figsize=(8,6))

	### Get the entropy values for selected generations
	gens = [1, 5, 10, 25, 50, 100, 150, 195, 199]
	results = utils.genpareto_entropy(gens, df=df)

	### Parse results for plotting both total pop and non-dom indvs
	col_ent = []
	row_ent = []
	ent_pos = []
	for key,val in results[0].items():
	    col_ent.append(val['col_entropy'])
	    row_ent.append(val['row_entropy'])
	    ent_pos.append(key)

	col_ent2 = []
	row_ent2 = []
	ent_pos2 = []
	for key,val in results[1].items():
	    col_ent2.append(val['col_entropy'])
	    row_ent2.append(val['row_entropy'])
	    ent_pos2.append(key)

	### Plot column-wise
	axes[0].violinplot(col_ent2, np.arange(len(col_ent)), 
		showextrema=False, bw_method=1)

	parts = axes[0].violinplot(col_ent, np.arange(len(col_ent)))   
	
	for pc in parts['bodies']:
	    pc.set_edgecolor('black')
	    pc.set_hatch('///')

	### Set ticks and tick labels
	axes[0].set_xticks(np.arange(len(ent_pos)))
	axes[0].set_xticklabels(ent_pos)

	### Set axis labels
	axes[0].set_xlabel('Generation')
	axes[0].set_ylabel('Shannon Entropy')

	### Create column-wise legend
	axes[0].legend(['All Individuals', 'Pareto Optimal\n Individuals'])
	legend =  axes[0].get_legend()
	legend.legendHandles[1].set_color('blue')
	legend.legendHandles[0].set_color('orange')
	legend.legendHandles[0].set_hatch('///')
	legend.legendHandles[0].set_alpha(.75)
	legend.legendHandles[0].set_edgecolor('k')

	### Plot row-wise
	axes[1].violinplot(row_ent2, np.arange(len(row_ent2)), 
	 	showextrema=False, bw_method=1)
	
	parts = axes[1].violinplot(row_ent, np.arange(len(row_ent)))
	for pc in parts['bodies']:
	    pc.set_edgecolor('black')
	    pc.set_hatch('///')

	### Set ticks and tick labels
	axes[1].set_xticks(np.arange(len(ent_pos)))
	axes[1].set_xticklabels(ent_pos)

	### Create row-wise legend
	axes[1].legend(['All Individuals','Pareto Optimal\n Individuals'])
	leg = axes[1].get_legend()
	leg.legendHandles[1].set_color('blue')
	leg.legendHandles[0].set_color('orange')
	leg.legendHandles[0].set_hatch('///')
	leg.legendHandles[0].set_alpha(.75)
	leg.legendHandles[0].set_edgecolor('k')

	### Set axis labels
	axes[1].set_xlabel('Generation')
	axes[1].set_ylabel('')

	### Save figure
	file = fn + '.png'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight', dpi=300)

	### Open figure
	time.sleep(0.5)
	os.system('open ' + uri)

	return None


def kde_plot(df, fn, output_dir):

	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(figsize=(8,6))

	### Sample dataframe
	df_sample = df.sample(1000000)

	### Plot data
	g = sb.jointplot(x='COST', y='AVOIDEDCOST_EW0', data=df_sample, kind='kde', 
		color='r', annot_kws=dict(facecolor='w',fontsize=0, edgecolor='w'))
	g.plot_joint(plt.scatter, c='k', s=30, linewidth=0.2, marker="+")
	g.ax_joint.collections[0].set_alpha(0)

	### Axis labels and ticks
	plt.xlabel('Restoration cost ($)', labelpad=10)
	plt.ylabel('PV of avoided damages ($)', labelpad=10)

	### Format x-tick labels
	xticks = [0, 1*10**6, 2*10**6, 3*10**6, 4*10**6]
	xlabels = []
	for i, tick in enumerate(xticks):
		xlabels.append('${:,.0f}'.format(tick))
	plt.xticks(xticks, xlabels)

	### Add y-tick labels
	yticks = [-400000, -200000, 0, 200000, 400000]
	ylabels = []
	for i, tick in enumerate(yticks):
		ylabels.append('${:,.0f}'.format(tick))
	plt.yticks(yticks, ylabels)

	### Save figure
	file = fn + '.jpg'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight', dpi=300)

	### Open figure
	time.sleep(0.5)
	os.system('open ' + uri)

	return None


def benefitcost_frontier(df, fn, output_dir, non_dominated=True, open_fig=True):

	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(figsize=(8,6))

	### Get data
	df = df[df['COST']<79100]

	x = np.array(df['COST']) # Restoration cost
	y0 = np.array(df['AVOIDEDCOST_EW0']) # Benefits (n=0)
	y1 = np.array(df['AVOIDEDCOST_ADJUSTED']) # Benefits (n=0)

	if non_dominated == True:
		# Plot dominated solutions
		axes.plot(x, y, 'ko', mfc='none', mew=.2, markersize=4, 
				  label='Sub-optimal scenarios')

	### Plot ROI=5 line
	roi_x = np.array(range(-10**6,10**6))
	roi_y = roi_x*5
	axes.plot(roi_x, roi_y/5, 'k--', alpha=0.7, label='Benefit:Cost = 1.0')
	axes.plot(roi_x, roi_y, 'k:', alpha=0.7, label='Benefit:Cost = 5.0')

	### Find non-dominated solutions
	pfront_x, pfront_y = utils.pareto_frontier(x, y1, max_x=False, max_y=True)
	# axes.plot(pfront_x, pfront_y, 'k-', label='Pareto frontier')
	axes.plot(pfront_x, pfront_y, 'bo-', label='Optimal scenarios\n(benefits capped)')
	print pd.DataFrame({'Cost': np.array(pfront_x), 
						'Benefit': np.array(pfront_y)})	

	pfront_x, pfront_y = utils.pareto_frontier(x, y0, max_x=False, max_y=True)
	# axes.plot(pfront_x, pfront_y, 'k-', label='Pareto frontier')
	axes.plot(pfront_x, pfront_y, 'ro-', label='Optimal scenarios')
	print pd.DataFrame({'Cost': np.array(pfront_x), 
						'Benefit': np.array(pfront_y)})

	### Set axis limits
	axes.set_xlim(-5000, 79000)
	axes.set_ylim(-10000, 410000)

	### Set axis labels
	xlabel = 'Restoration cost'
	ylabel = 'PV of avoided damages'
	axes.set_xlabel(xlabel, labelpad=10)
	axes.set_ylabel(ylabel, labelpad=10)

	### Format tick tick labels
	xlabels = range(0,100000,25000)
	axes.set_xticks(xlabels)
	for i, label in enumerate(xlabels):
		xlabels[i] = '${:,.0f}'.format(label)
	axes.set_xticklabels(xlabels)

	### Get total baseline damages costs
	base_dict = json.load(open(base_json_uri))
	total_damages = base_dict['DAMAGECOST_EW0']

	### Add percentages to y-labels
	ylabels = range(0,500000,100000)
	axes.set_yticks(ylabels)
	for i, label in enumerate(ylabels):
		ylabels[i] = '${:,.0f}'.format(label) 
		percent = round((label / total_damages) * 100, 1)
		ylabels[i] = '${:,.0f}'.format(label)  + ' (%s)' %(str(percent)+'%') 
	axes.set_yticklabels(ylabels)

	### Add legend
	handles, labels = axes.get_legend_handles_labels()
	legend = axes.legend(handles[::-1], labels[::-1], loc='upper left', frameon=True)	
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')
	legend.get_frame().set_facecolor('w')
	legend.get_frame().set_alpha(1)

	### Save figure
	file = fn + '.jpg'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight', dpi=300)
	
	### Open figure
	if open_fig:
		time.sleep(0.5)
		os.system('open ' + uri)

	return None


def benefitcost_maps(scenario_list, fn, output_dir):
	
	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(1, 3, figsize=(7.5,3))
	fig.tight_layout()
	plt.subplots_adjust(left=None, bottom=0, right=None, top=0.8, wspace=0.08)

	### Read shapefiles
	ws = 'lewiscreek'
	ws_shp = utils.shp(shp_uri_dict[ws+'_shp_uri'])
	xs_shp = utils.shp(shp_uri_dict[ws+'_xs_shp_uri'])
	stream_shp = utils.shp(shp_uri_dict[ws+'_stream_shp_uri'])	

	### Set plot bounding box
	bbox = ws_shp.bbox

	### Loop through
	for i, scenario in enumerate(scenario_list):

		scenario = map(int, scenario.values[0][1:-1].split(','))

		### Plot ws boundary
		ws_df = utils.shp.parse_shp(ws_shp)
		shape = ws_df['paths'][0]
		shape.set_edgecolor('k')
		shape.set_linewidth(0.8)
		axes[i].add_collection(shape)

		### Plot stream
		stream_df = utils.shp.parse_shp(stream_shp)
		for stream in range(len(stream_df)):	
			shape = stream_df['paths'][stream]
			shape.set_edgecolor('k')
			shape.set_linewidth(0.6)
			axes[i].add_collection(shape)

		### Plot all cross-sections
		xs_df = utils.shp.parse_shp(xs_shp)
		xs_df = xs_df.reset_index()
		for xs in range(len(xs_df)):
			shape = xs_df['paths'][xs]
			unrestored_color = (0, 0, 0, 0.4)
			shape.set_facecolor(unrestored_color)
			shape.set_linewidth(0)
			axes[i].add_collection(shape)		
		
		### Get xs_indices_df
		xs_indices_uri = os.path.join(paths.xs_dir, 'lewiscreek_xs_indices.csv')
		xs_indices_df = pd.read_csv(xs_indices_uri)
		xs_indices_df = xs_indices_df[xs_indices_df['DECISION_UNIT']==1]
		xs_indices_df = xs_indices_df.sort_values(['INDEX']).reset_index()

		### Define shape color and linewidth
		restoredforest_color = 'g'
		restoredwetland_color = 'b'
		reconnectivity_color = 'purple'

		def add_shape(xs, color):
			shape = xs_df['paths'][xs]
			shape.set_linewidth(0.1)
			shape.set_edgecolor('w')
			shape.set_facecolor(color)	
			return shape

		### Iterate through scenario data
		print
		print i
		for j, xs_code in enumerate(scenario):
			rs = xs_indices_df['STATION'][j]
			rs = float(rs)
			xs = xs_df.index[pd.to_numeric(xs_df['RiverStat'])==rs]
			xs = int(xs[0])
			if xs_code > 0:
				print j, rs, xs_code
			
			if xs_code == 1: 
				s = add_shape(xs, restoredforest_color)
			if xs_code == 2:
				s = add_shape(xs, restoredwetland_color)
			if xs_code == 3:
				s = add_shape(xs, reconnectivity_color)
		
		### Set axes' limits
		axes[i].set_xlim([bbox[0]-bbox[0]*.001, bbox[2]+bbox[2]*.001])
		axes[i].set_ylim([bbox[1]-bbox[0]*.001, bbox[3]+bbox[3]*.001])

		### Don't show ticks
		axes[i].xaxis.set_ticks_position('none')
		axes[i].yaxis.set_ticks_position('none')
		axes[i].get_xaxis().set_ticks([])
		axes[i].get_yaxis().set_ticks([])

	### Sub-plot titles
	axes[0].set_title('$25,000')
	axes[1].set_title('$50,000')
	axes[2].set_title('$75,000')

	### Add legend
	def make_proxy(color, lw):
		return Rectangle((0,0), 1.2, 0.5, color=color)

	proxies = []
	proxies.append(make_proxy(unrestored_color, 1.5))
	proxies.append(make_proxy(restoredforest_color, 1.5))
	proxies.append(make_proxy(restoredwetland_color, 1.5))
	proxies.append(make_proxy(reconnectivity_color, 1.5))

	labels = ['Unrestored', 
			  'Forest\nrevegetation', 
			  'Wetland\nrevegetation', 
			  'Floodplain lowering &\nwetland revegetation']
	legend = axes[1].legend(proxies, labels, loc='lower center', frameon=True,
							bbox_to_anchor=(0.5, -0.23), ncol=4)
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')

	### Save figure
	file = fn + '.jpg'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight', dpi=600, 
		frameon=False, transparent=True)
	
	### Open figure
	time.sleep(0.5)
	os.system('open ' + uri)


def benefitcost_distributionplot_baseline(fn, output_dir):

	### Get data
	stem = paths.optimization_output_dir
	csv_uri = os.path.join(stem, 'intervention_logbooks/baseline.csv')		
	df = pd.read_csv(csv_uri)
	df = df[~pd.isnull(df['PROPERTY_VALUE'])]
	df = df.copy()
	
	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(figsize=(7,5))

	### Plot guidelines 
	percent_reduction = 10
	x_1percent = np.linspace(0, 5000000, 1000)
	y_1percent = x_1percent / percent_reduction
	axes.plot(x_1percent, y_1percent, 'k-', alpha=.5, 
		label='%d%% of property value' %(100/percent_reduction))

	percent_reduction = 1
	x_1percent = np.linspace(0, 5000000, 1000)
	y_1percent = x_1percent / percent_reduction
	axes.plot(x_1percent, y_1percent, 'k--', alpha=.5, 
		label='%d%% of property value' %(100/percent_reduction))

	percent_reduction = 0.2
	x_1percent = np.linspace(0, 5000000, 1000)
	y_1percent = x_1percent / percent_reduction
	axes.plot(x_1percent, y_1percent, 'k:', alpha=.5, 
		label='%d%% of property value' %(100/percent_reduction))

	### Read in depth-damage functions from CSV & 
	### create columns for slope values
	ddf_df = pd.read_csv(depthdamage_funcs_uri)
	dd_funcs = list(ddf_df.columns)
	dd_funcs.remove('DEPTH_FT')
	for func in dd_funcs:
		if func != 'DEPTH_FT':
			ddf_df[func + '_slope'] = ddf_df[func].shift(-1) - ddf_df[func]

	### Get data of interest
	flood_events = [2, 10, 25, 100]
	for event in flood_events:
		### Calculate percent damage to properties based on inundation depth
		df = ras_postprocessing.depth2damage(df, ddf_df, event)

		### Calculate monetary damage to properties based on percent damage
		df = ras_postprocessing.percdamage2dollardamage(df, event)

	### Calculate expected annual damages using numerical integration of 
	### monetary damages from flood events and annual probability
	df = ras_postprocessing.trapezoidal_integration(df)
	
	### Calculate present value of expected annual damages over 100 year
	### time horizon and 7% discount rate
	df = ras_postprocessing.present_value(df)	

	print 'Total baseline damage costs = $', np.sum(df['PV'])

	bottom25 = np.percentile(df['PROPERTY_VALUE'], 25)
	top25 = np.percentile(df['PROPERTY_VALUE'], 75)

	damagecost_bottom25 = np.sum(df['PV'][df['PROPERTY_VALUE']<=bottom25])
	damagecost_top25 = np.sum(df['PV'][df['PROPERTY_VALUE']>=top25])
	print 'Properties in bottom quartile exposed to flooding: ', damagecost_bottom25
	print 'Properties in top quartile exposed to flooding: ',  damagecost_top25
	print
	# print ('Number of properties damaged by flooding:',
		# len(df[(df['PV'] / df['PROPERTY_VALUE'])>1]))
	print 'Number of properties damaged by flooding:', len(df[df['PV']>0])
	print len(df[df['PV']>=df['PROPERTY_VALUE']])
	print len(df[df['PV']>=(5*df['PROPERTY_VALUE'])])

	property_types = ['MOBILE HOME', 'SINGLE FAMILY DWELLING', 'COMMERCIAL']
	for property_type in property_types:

		if property_type == 'SINGLE FAMILY DWELLING':
			symbol = 'bo'
			label = 'Single-family home'
		if property_type == 'MOBILE HOME':
			symbol = 'ro'
			label = 'Mobile home'
		if property_type in ['COMMERCIAL', 'OTHER COMMERCIAL']:
			symbol = 'ko'
			label = 'Commercial'

		df_propertytype = df[df['SITETYPE']==property_type]

		### Plot properties with damage costs above zero
		df_abovezero = df_propertytype[df_propertytype['PV']!=0]
		df_abovezero = df_abovezero.sort_values(['PROPERTY_VALUE'])
		property_values = df_abovezero['PROPERTY_VALUE']
		damage_costs = df_abovezero['PV']
		axes.plot(property_values, damage_costs, symbol, label=label)
		print 
		print property_type
		print len(damage_costs)

		# axes[1].plot(property_values, damage_costs / property_values, symbol)

		### Plot properties with no damages
		df_zero = df_propertytype[df_propertytype['PV']==0]
		df_zero = df_zero.sort_values(['PROPERTY_VALUE'])
		property_values = df_zero['PROPERTY_VALUE']
		damage_costs = df_zero['PV']
		jitter = np.random.normal(loc=0, scale=0.05, size=len(property_values))
		axes.plot(property_values, damage_costs + jitter, symbol, mfc='none', 
			mew=.5, markersize=4, label='_nolegend_')

		# axes[1].plot(property_values, damage_costs + jitter, symbol, mfc='none', 
		# 	mew=.5, markersize=4, label='_nolegend_')

	print
	print df['PV'][df['PV']>0] / df['PROPERTY_VALUE'][df['PV']>0]
	print
	print np.mean(df['PV'][df['PV']>0] / df['PROPERTY_VALUE'][df['PV']>0])
	print np.median(df['PV'][df['PV']>0] / df['PROPERTY_VALUE'][df['PV']>0])

	### Set axis labels
	axes.set_xlabel('Property value ($)')
	axes.set_ylabel('PV of property damages ($)')

	# axes[1].set_xlabel('Property value ($)')
	# axes[1].set_ylabel(r'PV of property damages ($) : Property value ($)')

	### Set axes limits
	axes.set_xlim(9000, 3000000)
	axes.set_ylim(-0.5, 5000000)

	### Set axis scale
	axes.set_xscale('log')
	axes.set_yscale('symlog')

	# axes[1].set_xscale('log')

	### Formet tick labels
	fmt = '${x:,.0f}'
	axes.set_xticks([10000,  25000,  50000,   100000, 
					 250000, 500000, 1000000, 2500000])
	tick = ticker.StrMethodFormatter(fmt)
	axes.xaxis.set_major_formatter(tick) 
	axes.yaxis.set_major_formatter(tick) 

	# axes[1].set_xticks([10000,  25000,  50000,   100000, 
					 # 250000, 500000, 1000000, 2500000])
	# axes[1].xaxis.set_major_formatter(tick) 

	plt.minorticks_off()

	### Legend
	print 

	handles, labels = axes.get_legend_handles_labels()
	handles = [handles[3], handles[4], handles[5], handles[0], handles[1], handles[2]]
	labels = [labels[3], labels[4], labels[5], labels[0], labels[1], labels[2]]
	legend = plt.legend(handles, labels, frameon=True, bbox_to_anchor=(0.57, 0.1))
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')

	### Save figure
	file = fn + '.pdf'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight', transparent=True)
	
	### Open figure
	time.sleep(1)
	os.system('open ' + uri)


def ew_maps(scenario_list, n_list, fn, output_dir):
	
	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(1, 3, figsize=(7.5,3))
	fig.tight_layout()
	plt.subplots_adjust(left=None, bottom=0, right=None, top=0.8, wspace=0.08)

	### Read shapefiles
	ws = 'lewiscreek'
	ws_shp = utils.shp(shp_uri_dict[ws+'_shp_uri'])
	xs_shp = utils.shp(shp_uri_dict[ws+'_xs_shp_uri'])
	stream_shp = utils.shp(shp_uri_dict[ws+'_stream_shp_uri'])	

	### Set plot bounding box
	bbox = ws_shp.bbox

	### Loop through
	for i, scenario in enumerate(scenario_list):

		scenario = map(int, scenario.values[0][1:-1].split(','))

		### Plot ws boundary
		ws_df = utils.shp.parse_shp(ws_shp)
		shape = ws_df['paths'][0]
		shape.set_edgecolor('k')
		shape.set_linewidth(0.8)
		axes[i].add_collection(shape)

		### Plot stream
		stream_df = utils.shp.parse_shp(stream_shp)
		for stream in range(len(stream_df)):	
			shape = stream_df['paths'][stream]
			shape.set_edgecolor('k')
			shape.set_linewidth(0.6)
			axes[i].add_collection(shape)

		### Plot all cross-sections
		xs_df = utils.shp.parse_shp(xs_shp)
		xs_df = xs_df.reset_index()
		for xs in range(len(xs_df)):
			shape = xs_df['paths'][xs]
			unrestored_color = (0, 0, 0, 0.4)
			shape.set_facecolor(unrestored_color)
			shape.set_linewidth(0)
			axes[i].add_collection(shape)		
		
		### Get xs_indices_df
		xs_indices_uri = os.path.join(paths.xs_dir, 'lewiscreek_xs_indices.csv')
		xs_indices_df = pd.read_csv(xs_indices_uri)
		xs_indices_df = xs_indices_df[xs_indices_df['DECISION_UNIT']==1]
		xs_indices_df = xs_indices_df.sort_values(['INDEX']).reset_index()

		### Define shape color and linewidth
		restoredforest_color = 'g'
		restoredwetland_color = 'b'
		reconnectivity_color = 'purple'

		def add_shape(xs, color):
			shape = xs_df['paths'][xs]
			shape.set_linewidth(0.1)
			shape.set_edgecolor('w')
			shape.set_facecolor(color)	
			return shape

		### Iterate through scenario data
		print
		print i
		for j, xs_code in enumerate(scenario):
			rs = xs_indices_df['STATION'][j]
			rs = float(rs)
			xs = xs_df.index[pd.to_numeric(xs_df['RiverStat'])==rs]
			xs = int(xs[0])
			if xs_code > 0:
				print j, rs, xs_code

			if xs_code == 1: 
				s = add_shape(xs, restoredforest_color)
			if xs_code == 2:
				s = add_shape(xs, restoredwetland_color)
			if xs_code == 3:
				s = add_shape(xs, reconnectivity_color)
		
		### Set axes' limits
		axes[i].set_xlim([bbox[0]-bbox[0]*.001, bbox[2]+bbox[2]*.001])
		axes[i].set_ylim([bbox[1]-bbox[0]*.001, bbox[3]+bbox[3]*.001])

		### Don't show ticks
		axes[i].xaxis.set_ticks_position('none')
		axes[i].yaxis.set_ticks_position('none')
		axes[i].get_xaxis().set_ticks([])
		axes[i].get_yaxis().set_ticks([])

	### Sub-plot titles
	axes[0].set_title(r'$\eta$ = %s' %str(n_list[0]))
	axes[1].set_title(r'$\eta$ = %s' %str(n_list[1]))
	axes[2].set_title(r'$\eta$ = %s' %str(n_list[2]))

	### Add legend
	def make_proxy(color, lw):
		return Rectangle((0,0), 1.2, 0.5, color=color)

	proxies = []
	proxies.append(make_proxy(unrestored_color, 1.5))
	proxies.append(make_proxy(restoredforest_color, 1.5))
	proxies.append(make_proxy(restoredwetland_color, 1.5))
	proxies.append(make_proxy(reconnectivity_color, 1.5))

	labels = ['Unrestored', 
			  'Forest\nrevegetation', 
			  'Wetland\nrevegetation', 
			  'Floodplain lowering &\nwetland revegetation']
	legend = axes[1].legend(proxies, labels, loc='lower center', frameon=True,
							bbox_to_anchor=(0.5, -0.23), ncol=4)
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')

	### Save figure
	file = fn + '.jpg'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight', dpi=600, 
		frameon=False, transparent=True)
	
	### Open figure
	time.sleep(0.5)
	os.system('open ' + uri)

def ew_distributionplots(scenario_list, n_list, fn, output_dir):
		
	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(figsize=(7,5.5))

	### Plot guidelines 
	percent_reduction = 10
	x_percent = np.linspace(0, 5000000, 1000)
	y_percent = x_percent / percent_reduction
	axes.plot(x_percent, y_percent, 'k--', alpha=.5, 
		label='%s%% of property value' %str(100/percent_reduction))
	axes.plot(x_percent, [0]*len(x_percent), 'k-', 
		alpha=.5, label='No reduction in damage')

	### Get total baseline damages costs
	base_dict = json.load(open(base_json_uri))
	total_damagecost_base_pv = base_dict['DAMAGECOST_EW0']

	### Read in depth-damage functions from CSV & 
	### create columns for slope values
	ddf_df = pd.read_csv(depthdamage_funcs_uri)
	dd_funcs = list(ddf_df.columns)
	dd_funcs.remove('DEPTH_FT')
	for func in dd_funcs:
		if func != 'DEPTH_FT':
			ddf_df[func + '_slope'] = ddf_df[func].shift(-1) - ddf_df[func]

	### Get baseline damage costs
	csv_uri = os.path.join(intervention_logbooks_dir, 'baseline.csv')
	df = pd.read_csv(csv_uri)

	### Get data of interest
	flood_events = [2, 10, 25, 100]
	for event in flood_events:
		### Calculate percent damage to properties based 
		### on inundation depth
		df = ras_postprocessing.depth2damage(df, ddf_df, event)

		### Calculate monetary damage to properties based 
		### on percent damage
		df = ras_postprocessing.percdamage2dollardamage(df, event)

	### Calculate expected annual damages using numerical 
	### integration of monetary damages from flood events and 
	### annual probability
	df = ras_postprocessing.trapezoidal_integration(df)
	
	### Calculate present value of expected annual damages over 100 year
	### time horizon and 7% discount rate
	df = ras_postprocessing.present_value(df)
	damagecosts_base = df['PV']

	### Iterate through scenarios
	for i, scenario in enumerate(scenario_list):
		scenario = map(int, scenario.values[0][1:-1].split(','))
		
		### Get data
		avoided_damagecosts = pd.Series([0]*len(damagecosts_base))
		for xs, intervention in enumerate(scenario): 
			if intervention != 0:
				csv_uri = os.path.join(intervention_logbooks_dir, 
					'%s_%s.csv' %(str(xs), str(intervention)))
				df = pd.read_csv(csv_uri)

				### Get data of interest
				flood_events = [2, 10, 25, 100]
				for event in flood_events:
					### Calculate percent damage to properties based 
					### on inundation depth
					df = ras_postprocessing.depth2damage(df, ddf_df, event)

					### Calculate monetary damage to properties based 
					### on percent damage
					df = ras_postprocessing.percdamage2dollardamage(df, event)

				### Calculate expected annual damages using numerical 
				### integration of monetary damages from flood events and 
				### annual probability
				df = ras_postprocessing.trapezoidal_integration(df)
				
				### Calculate present value of expected annual damages over 100 year
				### time horizon and 7% discount rate
				df = ras_postprocessing.present_value(df)

				avoided_damagecosts += (damagecosts_base - df['PV'])

		property_values = df['PROPERTY_VALUE']
		x = property_values[(avoided_damagecosts>0)]
		y = avoided_damagecosts[(avoided_damagecosts>0)]
		ind = np.argsort(x)
		x = np.array(x)[ind]
		y = np.array(y)[ind]

		### Plot data
		if i == 0: color, eta = 'r', n_list[0]
		if i == 1: color, eta = 'k', n_list[1]
		if i == 2: color, eta = 'b', n_list[2]
		axes.plot(x, y, '%s-o' %color, label=r'$\eta$ = %s' %str(eta))

	### Set axis labels
	axes.set_xlabel('Property value ($)')
	axes.set_ylabel('Reduction in flood damages ($)')

	### Set axes limits
	axes.set_xlim(9000, 300000)
	axes.set_ylim(-0.9, 5000000)

	### Set axis scale
	axes.set_xscale('log')
	axes.set_yscale('symlog')

	### Formet tick labels
	fmt = '${x:,.0f}'
	axes.set_xticks([10000, 25000, 50000, 100000, 250000])
	tick = ticker.StrMethodFormatter(fmt)
	axes.xaxis.set_major_formatter(tick) 
	axes.yaxis.set_major_formatter(tick) 
	plt.minorticks_off()

	### Legend
	handles, labels = axes.get_legend_handles_labels()
	handles = [handles[2], handles[3], handles[4], handles[0], handles[1]]
	labels = [labels[2], labels[3], labels[4], labels[0], labels[1]]
	legend = plt.legend(handles, labels, loc='upper left', frameon=True)
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')

	### Save figure
	file = fn + '.pdf'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight', transparent=True)
	
	### Open figure
	time.sleep(1)
	# os.system('open ' + uri)


def ew_distributionbars(scenario_list, n_list, fn, output_dir):
		
	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(figsize=(7,5.5))

	### Get total baseline damages costs
	base_dict = json.load(open(base_json_uri))
	total_damagecost_base_pv = base_dict['DAMAGECOST_EW0']

	### Read in depth-damage functions from CSV & 
	### create columns for slope values
	ddf_df = pd.read_csv(depthdamage_funcs_uri)
	dd_funcs = list(ddf_df.columns)
	dd_funcs.remove('DEPTH_FT')
	for func in dd_funcs:
		if func != 'DEPTH_FT':
			ddf_df[func + '_slope'] = ddf_df[func].shift(-1) - ddf_df[func]

	### Get baseline damage costs
	csv_uri = os.path.join(intervention_logbooks_dir, 'baseline.csv')
	df = pd.read_csv(csv_uri)

	### Get data of interest
	flood_events = [2, 10, 25, 100]
	for event in flood_events:
		### Calculate percent damage to properties based 
		### on inundation depth
		df = ras_postprocessing.depth2damage(df, ddf_df, event)

		### Calculate monetary damage to properties based 
		### on percent damage
		df = ras_postprocessing.percdamage2dollardamage(df, event)

	### Calculate expected annual damages using numerical 
	### integration of monetary damages from flood events and 
	### annual probability
	df = ras_postprocessing.trapezoidal_integration(df)
	
	### Calculate present value of expected annual damages over 100 year
	### time horizon and 7% discount rate
	df = ras_postprocessing.present_value(df)
	damagecosts_base = np.array(df['PV'])

	# ### Iterate through scenarios
	for i, scenario in enumerate(scenario_list):
		scenario = map(int, scenario.values[0][1:-1].split(','))
		
		### Get data
		avoided_damagecosts = np.array([0]*len(damagecosts_base), dtype='float64')
		for xs, intervention in enumerate(scenario): 
			if intervention != 0:
				csv_uri = os.path.join(intervention_logbooks_dir, 
					'%s_%s.csv' %(str(xs), str(intervention)))
				df = pd.read_csv(csv_uri)

				### Get data of interest
				flood_events = [2, 10, 25, 100]
				for event in flood_events:
					### Calculate percent damage to properties based 
					### on inundation depth
					df = ras_postprocessing.depth2damage(df, ddf_df, event)

					### Calculate monetary damage to properties based 
					### on percent damage
					df = ras_postprocessing.percdamage2dollardamage(df, event)

				### Calculate expected annual damages using numerical 
				### integration of monetary damages from flood events and 
				### annual probability
				df = ras_postprocessing.trapezoidal_integration(df)
				
				### Calculate present value of expected annual damages over 
				### 100 year time horizon and 7% discount rate
				df = ras_postprocessing.present_value(df)

				avoided_damagecosts += (damagecosts_base - np.array(df['PV']))


		property_value = np.array(df['PROPERTY_VALUE'])
		property_value = property_value[avoided_damagecosts>0]
		avoided_damagecosts = avoided_damagecosts[avoided_damagecosts>0]

		bottom25 = np.nanpercentile(property_value, 25)
		median 	 = np.nanpercentile(property_value, 50)
		top25 	 = np.nanpercentile(property_value, 75)

		y1 = np.nansum(np.where(property_value<=bottom25, avoided_damagecosts, 0))
		y2 = np.nansum(np.where((property_value> bottom25) & 
			(property_value<=median), avoided_damagecosts, 0))
		y3 = np.nansum(np.where((property_value> median) & 
			(property_value<=top25), avoided_damagecosts, 0))
		y4 = np.nansum(np.where(property_value>=top25, avoided_damagecosts, 0))

		### Plot data
		if i == 0: color, eta, symbol = 'r', n_list[0], '^'
		if i == 1: color, eta, symbol = 'k', n_list[1], 's'
		if i == 2: color, eta, symbol = 'b', n_list[2], 'o'
		x = np.array([1,  2,  3,  4])
		y = np.array([y1, y2, y3, y4])
		width = 0.25
		axes.plot(x+width*i, y, symbol, mec='k', mfc='none', mew=1,
			label=r'$\eta$ = %s' %str(eta))

		print eta
		print y[0]
		print y[1]
		print y[2]
		print y[3]
		print 

	### Plot gridlines
	# axes.plot([1-width,1-width], [0,10**6], 'k--', alpha=0.5)
	axes.plot([2-width,2-width], [-1*10**8,10**8], 'k--', alpha=0.5)
	axes.plot([3-width,3-width], [-1*10**8,10**8], 'k--', alpha=0.5)
	axes.plot([4-width,4-width], [-1*10**8,10**8], 'k--', alpha=0.5)
	# axes.plot([5-width,5-width], [0,10**6], 'k--', alpha=0.5)


	### Set the position of the x ticks
	axes.set_xticks([p + width for p in x])

	### Set the labels for the x ticks
	axes.set_xticklabels(['First\nquartile', 'Second\nquartile', 
						  'Third\nquartile', 'Fourth\nquartile'])

	### Set axis limits
	axes.set_xlim(0.5+width, 5-width)
	axes.set_ylim(9000, 400000)

	### Transform axes
	# axes.set_yscale('symlog')
	axes.set_yscale('log')

	### Set y ticks
	axes.set_yticks([10000, 25000, 50000, 100000, 250000])

	### Formet tick labels
	fmt = '${x:,.0f}'
	tick = ticker.StrMethodFormatter(fmt)
	axes.yaxis.set_major_formatter(tick) 
	plt.minorticks_off()

	plt.tick_params(axis='x', bottom=False, pad=2)

	### Set axis labels
	axes.set_xlabel('Property value', labelpad=10)
	axes.set_ylabel('Reduction in flood damages ($)', labelpad=10)	
	
	### Legend
	handles, labels = axes.get_legend_handles_labels()
	legend = plt.legend(handles, labels, loc='upper left', frameon=True)
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')
	
	### Save figure
	file = fn + '.pdf'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight', transparent=True)
	
	### Open figure
	# time.sleep(1)
	# os.system('open ' + uri)


def equityweighted_benefitcost_fontier(df, fn, output_dir, open_fig=True):

	### Set plot parameters and style
	sb.set(style='ticks')
	fig, axes = plt.subplots(figsize=(8,6))

	### Get data
	df = df[df['FITNESS5']<500000]
	df = df.sort_values('FITNESS5')
	x = np.array(df['FITNESS5']) # Restoration cost

	y0 = np.array(df['FITNESS0'])
	y1 = np.array(df['FITNESS1'])
	y2 = np.array(df['FITNESS2'])
	y3 = np.array(df['FITNESS3'])
	y4 = np.array(df['FITNESS4'])

	isefficient_ew_neg2 = utils.is_pareto_efficient_indexed(np.column_stack((x,y0*-1)))
	isefficient_ew_neg1 = utils.is_pareto_efficient_indexed(np.column_stack((x,y1*-1)))
	isefficient_ew_0 = utils.is_pareto_efficient_indexed(np.column_stack((x,y2*-1)))
	isefficient_ew_1 = utils.is_pareto_efficient_indexed(np.column_stack((x,y3*-1)))
	isefficient_ew_2 = utils.is_pareto_efficient_indexed(np.column_stack((x,y4*-1)))


	# axes.plot(x[isefficient_ew_neg2], y2[isefficient_ew_neg2], 'k-')
	# axes.plot(x[isefficient_ew_neg2], y2[isefficient_ew_neg2], 'ro', 
		# label='Optimal scenarios - EW=-2')

	axes.plot(x[isefficient_ew_neg1], y2[isefficient_ew_neg1], 'k-')
	axes.plot(x[isefficient_ew_neg1], y2[isefficient_ew_neg1], 'go', 
		label='Optimal scenarios - EW=-1')

	axes.plot(x[isefficient_ew_0], y2[isefficient_ew_0], 'k-')
	axes.plot(x[isefficient_ew_0], y2[isefficient_ew_0], 'ko', 
		label='Optimal scenarios - EW=0')

	axes.plot(x[isefficient_ew_1], y2[isefficient_ew_1], 'k-')
	axes.plot(x[isefficient_ew_1], y2[isefficient_ew_1], 'co', 
		label='Optimal scenarios - EW=1')

	# axes.plot(x[isefficient_ew_2], y2[isefficient_ew_2], 'k-')
	# axes.plot(x[isefficient_ew_2], y2[isefficient_ew_2], 'mo', 
		# label='Optimal scenarios - EW=2')

	### Set axis limits
	axes.set_xlim(-10000, 450000)
	axes.set_ylim(-10000, 450000)

	### Set axis labels
	xlabel = 'Restoration cost ($)'
	ylabel = 'Present value of avoided damages ($)'
	axes.set_xlabel(xlabel, labelpad=10)
	axes.set_ylabel(ylabel, labelpad=10)

	### Format tick tick labels
	xlabels = axes.get_xticks().tolist()
	for i, label in enumerate(xlabels):
		xlabels[i] = '${:,.0f}'.format(label)
	axes.set_xticklabels(xlabels)

	### Get total baseline damages costs
	base_dict = json.load(open(base_json_uri))
	total_damages = base_dict['DAMAGECOST_EW0']

	### Add percentages to y-labels
	ylabels = axes.get_yticks().tolist()
	for i, label in enumerate(ylabels):
		ylabels[i] = '${:,.0f}'.format(label) 
		percent = round((label / total_damages) * 100, 2)
		ylabels[i] = '${:,.0f}'.format(label)  + ' (%s)' %(str(percent)+'%') 
	axes.set_yticklabels(ylabels)

	### Add legend
	handles, labels = axes.get_legend_handles_labels()
	handles = handles[::-1]
	labels = labels[::-1]

	legend = axes.legend(handles, labels, loc='upper left', frameon=True)	
	legend.get_frame().set_linewidth(1)
	legend.get_frame().set_edgecolor('k')

	### Save figure
	file = fn + '.png'
	uri = os.path.join(output_dir, file)
	plt.savefig(uri, bbox_inches='tight', dpi=300)
	
	### Open figure
	if open_fig:
		time.sleep(0.5)
		os.system('open ' + uri)

	return None