import os
import deap
import time
import random
import numpy as np
import pandas as pd

from deap import base, creator, tools, algorithms

def optimize(args):
	"""Genetic algorithm.

	Parameters:
	    args (dict): dictionary containing GA parameters

	Returns:
	    logbook (object): contains GA summary statistics
	"""

	### Initialization
	toolbox = init_toobox()
	toolbox = init_multiprocessing(toolbox, args)
	toolbox, pop = init_pop(toolbox, args)
	toolbox = init_mutation(toolbox, args)
	toolbox = init_crossover(toolbox, args)
	toolbox = init_selection(toolbox, args)
	toolbox = init_evaluate(toolbox, args)
	stats = init_stats()
	logbook = init_logbook()

	### Algorithm
	logbook = algorithm(toolbox, args, pop, stats, logbook)

	return logbook


def init_toobox():
	"""Initialize toolbox.

	Parameters:
		None

	Returns:
		toolbox (object): DEAP toolbox object
	"""

	toolbox = base.Toolbox()

	return toolbox


def init_multiprocessing(toolbox, args):
	"""Register multiple processors.
	
	Parameters:
		toolbox (object): DEAP toolbox object
		args (dict): dictionary containing GA parameters

	Returns:
		toolbox (object): DEAP toolbox object
	"""

	if args['multiprocessing'] == 'scoop':
		from scoop import futures
		toolbox.register("map", futures.map)

	elif args['multiprocessing'] == 'multi':
		import multiprocessing
		pool = multiprocessing.Pool()
		toolbox.register("map", pool.map)

	else:
		pass

	return toolbox


def init_pop(toolbox, args):
	"""Create fitness class for maximization problem and initialize population 
	with individuals with attributes.

	Parameters:
		toolbox (object): DEAP toolbox object
		args (dict): dictionary containing GA parameters

	Returns:
		toolbox (object): DEAP toolbox object
		pop (list): list of individuals 
	"""

	### Create Individual class that is derived from a list
	weights = args['weights']
	creator.create("FitnessMulti", base.Fitness, weights=weights)
	creator.create("Individual", list, fitness=creator.FitnessMulti)

	def init_indv(icls, indv_list):
		return icls(indv_list)
	toolbox.register("individual", init_indv, creator.Individual)

	pop = []
	pr_list = [0, 0.2, 0.4, 0.6, 0.8, 1.0]
	for p1 in pr_list:
		for p2 in pr_list:
			for p3 in pr_list:

				if (p1 + p2 + p3) > 1:
					pass
				
				else:
					pcum_list = [p1, p1+p2, p1+p2+p3]

					for indv in range(args['pop_size']/56):
						indv = [0] * args['indv_length']

						for attr in range(len(indv)):
							rand = random.random()

							if rand <= pcum_list[0]:
								indv[attr] = 1
							if (rand > pcum_list[0]) and (rand <= pcum_list[1]):
								indv[attr] = 2
							if (rand > pcum_list[1]) and (rand <= pcum_list[2]):
								indv[attr] = 3

						indv = toolbox.individual(indv)
						pop.append(indv)

	return toolbox, pop


def init_mutation(toolbox, args):
	"""Register mutation operator. 

	Parameters:
		toolbox (object): DEAP toolbox object
		args (dict): dictionary containing GA parameters

	Returns:
		toolbox (object): DEAP toolbox object
	"""

	n_interventions = args['N_interventions']

	if (args['mut_operator'] == 'mutFlipBit') or (n_interventions == 1):
		toolbox.register('mutate', tools.mutFlipBit, 
						 indpb=1/args['indv_length'])

	if args['mut_operator'] == 'mutUniformInt':
		toolbox.register('mutate', tools.mutUniformInt, low=0, 
						 up=n_interventions, indpb=1/args['indv_length'])

	return toolbox


def init_crossover(toolbox, args):
	"""Register crossover (xo) operator.

	Parameters:
		toolbox (object): DEAP toolbox object
		args (dict): dictionary containing GA parameters

	Returns:
		toolbox (object): DEAP toolbox object
	"""

	if args['cx_operator'] == 'cxOnePoint':
		toolbox.register('mate', tools.cxOnePoint) 

	if args['cx_operator'] == 'cxTwoPoint':
		toolbox.register('mate', tools.cxTwoPoint) 

	if args['cx_operator'] == 'cxUniform': 
		toolbox.register('mate', tools.cxUniform, indpb=0.5)

	return toolbox


def init_selection(toolbox, args):
	"""Register selection operator.

	Parameters:
		toolbox (object): DEAP toolbox object
		args (dict): dictionary containing GA parameters

	Returns:
		toolbox (object): DEAP toolbox object
	"""

	if args['sel_operator'] == 'selTournament':
		toolbox.register('select', tools.selTournament, tournsize=3)

	if args['sel_operator'] == 'selNSGA2':
		toolbox.register('select', tools.selNSGA2)

	if args['sel_operator'] == 'selSPEA2':
		toolbox.register('select', tools.selSPEA2)

	return toolbox


def init_evaluate(toolbox, args):
	"""Register evaluation operator. 

	Parameters:
		toolbox (object): DEAP toolbox object
		args (dict): dictionary containing GA parameters

	Returns:
		toolbox (object): DEAP toolbox object
	"""

	def evaluate(individual):
		"""Initialize objective function.
		
		Parameters:
			individual (list): individual in population, encoded as list of ints

		Returns:
	        fitness1 (float): 1st objective fitness value
	        fitness2 (float): 2nd objective fitness value
		"""

		fitness_func = args['fit_func']
		fitnesses = fitness_func(individual, args)

		return fitnesses


	toolbox.register('evaluate', evaluate)

	return toolbox


def init_stats():
	"""Register statistics.

	Parameters:
		None

	Returns:
		stats (object): GA statistics object
	"""

	stats_fit = tools.Statistics(key=lambda ind: ind.fitness.values)
	stats = tools.MultiStatistics(fitness=stats_fit)
	stats.register('mean', np.mean, axis=0)
	stats.register('std', np.std, axis=0)
	stats.register('min', np.min, axis=0)
	stats.register('max', np.max, axis=0)

	return stats


def init_logbook():
	"""Register logbook.
	
	Parameters:
		None

	Returns:
		logbook (object): GA logbook object
	"""

	logbook = tools.Logbook()

	return logbook


def algorithm(toolbox, args, pop, stats, logbook):
	"""Genetic algorithm, iterated over generations. Statistics for each 
		generation save to logbook.

	Parameters:
		toolbox (object): DEAP toolbox object
		args (dict): dictionary containing GA parameters
		pop (list): list of individuals 
		stats (object): GA statistics object
		logbook (object): GA logbook object

	Returns:
		logbook (object): GA logbook object
	"""

	### Evaluate initial population
	print '** EVALUATING INITIAL POPULATION **'

	invalid_indv = [indv for indv in pop if not indv.fitness.valid]
	fitnesses = toolbox.map(toolbox.evaluate, invalid_indv)
	for indv, fit in zip(invalid_indv, fitnesses):
		indv.fitness.values = fit

	### Iterate through generations
	for g in range(args['gens']):
		print '** STARTING GENERATION #: %s **' %str(g)
		start_time = time.time()

		### Set generation counter
		args['gen_counter'] += 1

		### Vary the pool of individuals
		print '		Generating offspring'
		offspring = algorithms.varOr(pop, toolbox, args['pop_size'], 
									 args['cxpb'], args['mutpb'])

		### Evaluate the individuals with an invalid fitness
		print '		Evaluating individuals'
		invalid_indv = [ind for ind in offspring if not ind.fitness.valid]
		fitnesses = toolbox.map(toolbox.evaluate, invalid_indv)
		for ind, fit in zip(invalid_indv, fitnesses):
			ind.fitness.values = fit

		### Select indiviudals for next generation
		print '		Selecting parents for next generation'
		pop[:] = toolbox.select(pop + offspring, args['pop_size'])

		### Calculate statistics and update logbook
		record = stats.compile(pop)
		logbook.record(gen=g + 1, evals=args['pop_size'], **record)

		### Print elapsed time
		elapsed_time = time.time() - start_time
		print '		Elapsed time: ' + time.strftime("%H:%M:%S", 
			time.gmtime(elapsed_time))
		print

	return logbook

