import os
import platform
import numpy as np
import pandas as pd

def evaluate_fitness(scenario, args):
	"""Genetic algorithm.

    Parameters:
    	scenario (list): encoding for individual restoration scenario
        args_dict (dict): dictionary containing objective function parameters

    Returns:
		fitnesses (list): list of objectives' fitnesses
    """

	### Initialize variables in args_dict
	ga_df = args['ga_df']
	base_dict = args['base_dict']
	obj_metrics = args['obj_metrics']	

	### Evaluate fitnesses
	fitnesses, metrics_dict = define_fitness(
		scenario, ga_df, base_dict, obj_metrics)
	
	### Export logbook
	args = update_summarylogbook(args, scenario, fitnesses, metrics_dict)

	return fitnesses


def define_fitness(scenario, ga_df, base_dict, obj_metrics):
	### Define fitness of objectives 
	scenario = pd.Series(scenario)
	xs = scenario.index[scenario!=0]
	ga_df_scenario = ga_df[ga_df['XS'].isin(xs)]
	ga_df_scenario= ga_df_scenario.sort_values(by=['XS']).reset_index(drop=True)
	intervention = scenario[scenario!=0].repeat(3).reset_index(drop=True)
	ga_df_scenario = ga_df_scenario[ga_df_scenario['INTERVENTION']==intervention]

	fitnesses = []
	for i, obj in enumerate(obj_metrics):
		if obj.split('_')[0] == 'avoidedcost':
			n = obj.split('_')[2]
			fitness = np.nansum(ga_df_scenario['AVOIDEDCOST_EW%s' %str(n)])
		if obj.split('_')[0] == 'restorationcost':
			fitness = np.nansum(ga_df_scenario['RESTORATION_COST'])

		fitnesses.append(fitness)

	metrics_dict = {}
	for n in [-2, -1, 0, 1, 2]:
		metrics_dict['AVOIDEDCOST_EW%s' %str(n)] = np.nansum(
			ga_df_scenario['AVOIDEDCOST_EW%s' %str(n)])

	metrics_dict['COST'] = np.nansum(ga_df_scenario['RESTORATION_COST'])


	return fitnesses, metrics_dict


def update_summarylogbook(args, scenario, fitnesses, metrics_dict):
	"""Update summary logbook csv with scenario outputs.

	Parameters:
		args (dict): objective function arguments dictionary
		counter (int): number of function evaluations
		scenario (list): encoding for individual restoration scenario
		fitnesses (list): list of objectives' fitnesses

	Returns:
		None
	"""

	### Read summary logbook to pandas dataframe
	summary_logbook_df = args['summary_logbook_df']

	### Put fitness values in dictionary
	fitnesses_dict = {}
	for i, fitness in enumerate(fitnesses):
		fitnesses_dict['FITNESS%s' %str(i)] = [fitness]

	### Create dictionary for all other outputs
	outputs_dict = {
		'GENERATION': [args['gen_counter']],
		'INDIVIDUAL': [scenario],
		'AVOIDEDCOST_EW-2': [metrics_dict['AVOIDEDCOST_EW-2']],
		'AVOIDEDCOST_EW-1': [metrics_dict['AVOIDEDCOST_EW-1']],
		'AVOIDEDCOST_EW0': [metrics_dict['AVOIDEDCOST_EW0']],
		'AVOIDEDCOST_EW1': [metrics_dict['AVOIDEDCOST_EW1']],
		'AVOIDEDCOST_EW2': [metrics_dict['AVOIDEDCOST_EW2']],
		'COST': [metrics_dict['COST']]
		}

	### Combine fitnesses_dict and outputs_dict and put in dataframe
	outputs_dict.update(fitnesses_dict)
	summary_outputs_df = pd.DataFrame(outputs_dict)

	### Concatenate existing summary_logbook_df and summary_outputs_df
	summary_logbook_df = pd.concat([summary_logbook_df, summary_outputs_df])
	args['summary_logbook_df'] = summary_logbook_df

	return args

