__author__ = 'Lucas Howard'
#This libarary contains the HECRAS_Calibration class which contains objects and methods for running a HEC-RAS model
#and determining its error with respect to calibration measurements as well as setting Manning's n values for all
#cross sections.


#Import numpy library for array, matrix and numerical operations and data types:
from numpy import *

#Import optimize package for finite difference gradient approximation
from scipy import optimize

#Import os libarary for running executables
import os



#Define HECRAS_Calibration class:
class HECRAS_Calibration:
    def __init__(self, location, name, N, h, Q):
        #location of the calibration directory as a string:
        self.location = location
        #list of n values as an array:
        self.N = N
        #list of calibration stages to be matched as an array:
        self.hcal = h
        #number of n values:
        self.nvars = len(N)
        #name of the HEC-RAS model as a string:
        self.name = name
        #stages predicted by HEC-RAS, preallocated as an array:
        self.h_pred = zeros(len(h))
        #the flow number as an integer (i.e., for calibration flow 1, Q=1, etc.)
        self.Q = Q
        #write the name of the model to a text file so that it can be run from the command line:
        f = open(location + r'\ModelName.txt', 'w')
        f.write(self.name)
        f.close()

    def SetManningsN(self):
        #Method that sets the Manning's n values as specified in self.N.  No variables are returned.

        #First open and close the ManningsN.txt file in write mode to delte its contents:
        thestr = r'\ManningsN.txt'
        thestr = self.location + thestr
        f = open(thestr, 'w')
        f.write("")
        f.close()
        #Write the values in self.N to the file, one value per line:
        with open(thestr, 'a') as f:
            for n in self.N:
                f.write(str(n) + "\n")
        #Move to the calibration directory so that the executable is looking in the right place for the model:
        os.chdir(self.location)
        #Run an executable that sets the n values in the HEC-RAS model and closes HEC-RAS
        os.system("SetMannings.exe")

    def GetStages(self):
        #Method that retrieves the simulated stages.

        #Read from the stage_output.txt file:
        thestr = r'\stage_output.txt'
        thestr = self.location + thestr
        with open(thestr, 'r') as f:
            i = 0
            for line in f:
                #For each line, convert to a floating point from a string and store in the self.h_pred array:
                self.h_pred[i] = float(line)
                i = i + 1

    def RunRas(self):
        #Method that runs a HEC-RAS simulation

        #Navigate to the appropriate directory and call the RunRAS.exe executable that performs a simulation
        os.chdir(self.location)
        os.system('RunRAS.exe')

    def zerofunction(self,N):
        retval=zerofunc(N,self)
        return(retval)
    def MeasureN(self,maxit):
        nvec=FindRoot(self.zerofunction,self.N,self,maxit)
        return(nvec)


def zerofunc(N,obj):
    #Standalone function, not a method of HECRAS_Calibration class.

    #Takes a list of Manning's n values given by the array N, and the HECRAS_Calibration object obj, and returns the
    #difference between the simulated and measured stage for a given flow, specified by Q.

    #Set N:
    obj.N = N
    obj.SetManningsN()

    #Perform the simulation:
    obj.RunRas()

    #Retrieve the stage data from the completed simulation:
    obj.GetStages()

    #Calcluate and return the error:
    retval = obj.h_pred[obj.Q-1]-obj.hcal[obj.Q-1]
    return retval


def FindRoot(func, x0, obj, maxiterations):
    #Root finding function for getting a "measurement" of the Manning's n values for a given piece of calibration data

    #Input variables:
    #func: the function whose zero is to be found
    #x0: the initial guess of the zero
    #obj: HECRAS_Calibration object, needed for estimating the gradient
    #maxiterations: the maximum number of iterations before exit


    #Apprximate the gradient with a stepsize equal to 1/10 the smallest paramter of the initial guess:
    epsilon=min(x0)/10
    grad = optimize.approx_fprime(x0, func, epsilon, obj)
    #set the iterations counter at 0:
    it_counter = 0

    #Initialize variables to track the x values:
    xlast = x0
    xnext = xlast
    #Calculate initial (currenct) function value:
    fvalc = func(x0, obj)
    #Set step size paramter to 1 (meaning the next guess is where the zero would be if the function were linear, smaller
    #beta values move somewhere between the current locatio and the projected linear zero)
    beta=1


    while (abs(fvalc) > .01) & (it_counter < maxiterations): #tolerance manually set to .01
        #approximate the gradient:
        epsilon=min(xnext)/10
        grad = optimize.approx_fprime(xnext, func, epsilon, obj)
        #Calculate next guess of zero using secant method:
        xnext = xlast - beta* grad * (fvalc / linalg.norm(grad))
        #increment the iteration counter:
        it_counter = it_counter + 1
        #Calculate the function value for the next guess:
        fvaln = func(xnext, obj)
        if abs(fvaln)<abs(fvalc): #if this value is closer to zero that the last one:
            #set last x value equal to next x value
            xlast = xnext
            #set beta to 1:
            beta=1
            #reset function values:
            fvalc=fvaln
        else:beta=.5*beta #else halve the step size and do not update function or x values, redo the iteration


        print(fvalc)
        print(it_counter)
    return (xnext)

def rms(N, obj):
    #Calculates the rms error of the model wrt calibration data as specified in the obj HECRAS_Calibration object
    obj.N = N
    obj.SetManningsN()
    obj.RunRas()
    obj.GetStages()
    retval = sqrt(sum((obj.h_pred - obj.hcal)**2))/sqrt(len(obj.hcal))
    return retval