import os
import numpy as np

from pygeoprocessing import geoprocessing, routing

import util


def execute(dem_uri, output_directory):
    """This function invokes a series of functions to calculate the compound 
    topographic index (CTI) for a digital elevation model (DEM).

    Parameters:
        dem_uri (string): path to unfilled DEM GeoTIFF raster
        output_directory (string): output directory for intermediate,
            temporary, and final files

    Returns:
        None
    """
    if not os.path.exists(output_directory):
    	os.makedirs(output_directory)

    if not os.path.exists(output_directory + 'intermediate/'):
        os.makedirs(output_directory + 'intermediate/')

    # Calculate flow direction
    flow_dir_uri = flow_dir(dem_uri, output_directory)

    # Calculate flow accumulation
    flow_acc_uri = flow_acc(flow_dir_uri, dem_uri, output_directory)

    # Calculate upland contributing area
    uca_uri = upland_contributing_area(flow_acc_uri, output_directory)

    # Calculate slope in percent
    slope_percent_uri = slope(dem_uri, output_directory)

    # Calculate slope in percent to slope in radians
    slope_radians_uri = slope_pct2radians(slope_percent_uri, output_directory)

    # Calculate compound topographic index
    cti_uri = compound_topographic_index(uca_uri, slope_radians_uri, 
                                                 output_directory)

    
    return None


def flow_dir(dem_uri, output_directory):
    """Route flux across DEM to guide flow from a d-infinity flow algorithm.
        This function will route flux across a landscape given a DEM to
        guide flow from a d-infinity flow algorithm, and a custom function
        that will operate on input flux and other user defined arguments
        to determine nodal output flux.

    Parameters:
        dem_uri (string): path to filled DEM GeoTIFF raster
        output_directory (string): output directory for intermediate,
            temporary, and final files

    Returns:
        flow_dir_uri (string): a uri to raster with d infinity flow 
            directions
    """
    print 'Calculating flow direction'
    
    flow_dir_uri = os.path.join(output_directory, 'intermediate/flow_dir.tif')
    routing.flow_direction_d_inf(dem_uri, flow_dir_uri)

    return flow_dir_uri


def flow_acc(flow_dir_uri, dem_uri, output_directory):
    """Calculate flow accumulation. A helper function to calculate flow 
        accumulation, also returns intermediate rasters for future 
        calculation.

    Parameters:
        flow_dir_uri (string): a path to raster with d infinity flow 
            directions
        dem_uri (string): a path to filled DEM GeoTIFF raster
        output_directory (string): output directory for intermediate,
            temporary, and final files

    Returns:
        flow_acc_uri (string): a path to raster with flow accumulation values
    """
    print 'Calculating flow accumulation'

    flow_acc_uri = os.path.join(output_directory, 'intermediate/flow_acc.tif')
    routing.flow_accumulation(flow_dir_uri, dem_uri, flow_acc_uri)

    return flow_acc_uri

def upland_contributing_area(flow_acc_uri, output_directory):
	"""Calculate the upland contributing area based on the flow accumulation 
		raster and its cell size.

	Parameters:
		flow_acc_uri (string): a path to raster with flow accumulation values
		output_directory (string): output directory for intermediate,
            temporary, and final files

    Returns:
    	uca_uri (string): a path to the upland contributing area raster
	
	"""
	print 'Calculating upland contributing area'

	uca_uri = os.path.join(output_directory, 'intermediate/uca.tif')
	flow_acc_array, flow_acc_gt = util.raster2array(flow_acc_uri)
    flow_acc_proj = util.get_raster_projection(flow_acc_uri)
	cellsize = flow_acc_gt[1]
	uca_array = (flow_acc_array + 1) * cellsize
	util.array2raster(uca_array, flow_acc_gt, flow_acc_proj, uca_uri)

	return uca_uri


def slope(dem_uri, output_directory):
	"""Create slope raster from DEM raster. Follows the algorithm described here:
    http://webhelp.esri.com/arcgiSDEsktop/9.3/index.cfm?TopicName=How%20Slope%20works

    Parameters:
        dem_uri (string): a URI to a  single band raster of z values.
        output_directory (string): output directory for intermediate,
            temporary, and final files

    Returns:
        slope_percent_uri (string): a path to the output slope uri in percent
    """
	print 'Calculating slope'

	slope_pct_uri = os.path.join(output_directory, 
		'intermediate/slope_percent.tif')
	geoprocessing.calculate_slope(dem_uri, slope_pct_uri)

	return slope_pct_uri


def slope_pct2radians(slope_pct_uri, output_directory):
	"""Convert slope raster in percent to slope raster in radians. 

	Parameters:
		slope_pct_uri (string): a path to the output slope uri in percent
		output_directory (string): output directory for intermediate,
            temporary, and final files

	Returns:
		slope_percent_uri (string): a path to the output slope uri in radians
	"""
	print 'Converting slope from percent to radians'

	slope_radians_uri = os.path.join(output_directory, 
		'intermediate/slope_radians.tif')
	slope_pct_array, slope_pct_gt = util.raster2array(slope_pct_uri)
    slope_pct_proj = util.get_raster_projection(slope_pct_uri)
	slope_radians_array = np.arctan(slope_pct_array / 100) 
	util.array2raster(slope_radians_array, slope_pct_gt, 
					  slope_pct_proj, slope_radians_uri)

	return slope_radians_uri


def compound_topographic_index(uca_uri, slope_radians_uri, output_directory):
	"""Calculates the compound topographic index as a function of the upland
		contributing area and the slope (in radians).

	Parameters:
		uca_uri (string): a path to the upland contributing area raster
		slope_radians_uri (string): a path to the slope (in radians) raster 
		output_directory (string): output directory for intermediate,
            temporary, and final files

	Returns: 
		cti_uri (string): a path to the output CTI raster
	"""
	print 'Calculating compound topographic index'

	cti_uri = os.path.join(output_directory, 'cti.tif')
	uca_array, uca_gt = util.raster2array(uca_uri)
    uca_proj = util.get_raster_projection(uca_uri)
	slope_array, slope_gt = util.raster2array(slope_radians_uri)
	tan_slope_array = np.where(slope_array > 0, 
							   np.tan(slope_array), 0.001)
	cti_array = np.log(uca_array / tan_slope_array)
	util.array2raster(cti_array, uca_gt, uca_proj, cti_uri)

	return cti_uri

