import os
import numpy as np
import pygeoprocessing as pygeo

from scipy.ndimage.measurements import label

import util, constants

def execute(cti_uri, lulc_uri, drainage_uri, wetlands_uri):
	"""This function invokes a series of functions to identify the locations
	of potential wetlands.

    Parameters:
		cti_uri (string): path to CTI GeoTIFF raster
		lulc_uri (string): path to LULC GeoTIFF raster
		drainage_uri (string): path to drainage GeoTIFF raster
		wetlands_uri (string): path to wetlands GeoTIFF raster

    Returns:
       None
    """

    # Create output directory
	output_directory = os.path.dirname(os.path.abspath(wetlands_uri))
	if not os.path.exists(output_directory):
		os.makedirs(output_directory)

	# Initialize constants
	cti_threshold = constants.cti_threshold
	nlcd_lulc_codes = constants.nlcd_lulc_codes
	gssurgo_drainage_codes = constants.gssurgo_drainage_codes
	wetland_area_threshold = constants.wetland_area_threshold

	# Execute functions
	cti_array, lulc_array, drainage_array, gt, proj = open_rasters(
		cti_uri, lulc_uri, drainage_uri)
	wetlands_array = identify_wetlands(cti_array, lulc_array, drainage_array,
		cti_threshold, nlcd_lulc_codes, gssurgo_drainage_codes)
	cti_array, lulc_array, drainage_array = None, None, None
	wetlands_array = subset_wetlands(wetlands_array, gt, wetland_area_threshold)
	wetlands_array2raster(wetlands_array, gt, proj, wetlands_uri)

	return None


def open_rasters(cti_uri, lulc_uri, drainage_uri):
	"""This function checks that all rasters are in the same projection, 
	converts them to numpy arrays, and resizes and resamples them them such 
	that they are aligned with the CTI array.

    Parameters:
		cti_uri (string): path to CTI GeoTIFF raster
		lulc_uri (string): path to LULC GeoTIFF raster
		drainage_uri (string): path to drainage GeoTIFF raster

    Returns:
       cti_array (numpy array): an array with CTI values
       lulc_array (numpy array): an array with LULC codes
       drainage_array (numpy array): an array with drainage codes
       cti_gt (python list): the geotransformation for the CTI raster
       cti_proj (string): the projection string for the CTI raster
    """

	print 'Opening rasters to arrays'
	
	# Check that all rasters are in the same projection
	datasets = [cti_uri, lulc_uri, drainage_uri]
	if not pygeo.assert_datasets_in_same_projection(datasets):
		print 'Datasets are not in same projection'
		quit()

	# Convert CTI to array and use as base dataset
	cti_array, cti_gt = util.raster2array(cti_uri)
	cti_proj = util.get_raster_projection(cti_uri)
	nrows, ncols = cti_array.shape
	cellsize = cti_gt[1]
	upper_left_x = cti_gt[0]
	upper_left_y = cti_gt[3]
	lower_right_x = upper_left_x + (cellsize * ncols)
	lower_right_y = upper_left_y - (cellsize * nrows)
	bbox = [upper_left_x, upper_left_y, lower_right_x, lower_right_y]

	# Resize and resample lulc_uri
	lulc_resampled_uri = pygeo.temporary_filename(suffix='.tif')
	pygeo.resize_and_resample_dataset_uri(lulc_uri, bbox, cellsize, 
										  lulc_resampled_uri, 'nearest')

	# Resize and resample drainage_uri
	drainage_resampled_uri = pygeo.temporary_filename(suffix='.tif')
	pygeo.resize_and_resample_dataset_uri(drainage_uri, bbox, cellsize, 
										  drainage_resampled_uri, 'nearest')

	# Convert resampled LULC and drainage rasters to arrays
	lulc_array, gt = util.raster2array(lulc_resampled_uri)
	drainage_array, gt = util.raster2array(drainage_resampled_uri)

	return cti_array, lulc_array, drainage_array, cti_gt, cti_proj


def identify_wetlands(cti_array, lulc_array, drainage_array, cti_threshold, 
					  lulc_codes, drainage_codes):
	"""This function identifies cells within a numpy array that have 
	characteristics of potential wetlands. It uses CTI, LULC, and drainage 
	datasets.

    Parameters:
		cti_array (numpy array): an array with CTI values
		lulc_array (numpy array): an array with LULC codes
		drainage_array (numpy array): an array with drainage codes
		cti_threshold (integer): the threshold value at which CTI may be 
			potentially be considered a wetland
		lulc_codes (python dictionary): a dictionary associating LULC codes
			with LULC classes
		drainage_codes (python_dictionary): a dictionary associating drainage 
			codes with drainage classes

    Returns:
        wetlands_array (numpy array): a binary array identifying potential 
        	wetlands 
    """
	
	print 'Identifying wetlands'

	# Create binary array where CTI array is greater than CTI threshold
	cti_binary_array = np.where(cti_array >= cti_threshold, 1, 0)
	
	# Create binary array where LULC is of certain classes listed below
	lulc_array = np.where((lulc_array == lulc_codes['Developed, Open Space']) |
						  (lulc_array == lulc_codes['Developed, Low Intensity']) |
						  (lulc_array == lulc_codes['Barren Land']) | 
						  (lulc_array == lulc_codes['Pasture/Hay']) | 
						  (lulc_array == lulc_codes['Cultivated Crops']) |
						  (lulc_array == lulc_codes['Woody Wetlands']) |
						  (lulc_array == lulc_codes['Emergent Herbaceous Wetlands']), 
						  1, 0)
	
	# Create binary array where drainage is poor
	drainage_array = np.where((drainage_array == drainage_codes['Very poorly drained']) |
							  (drainage_array == drainage_codes['Poorly drained']) |
							  (drainage_array == drainage_codes['Somewhat poorly drained']) |
							  (np.isnan(drainage_array)),
							  1, 0)
	
	# Create wetlands array where binary CTI, LULC, and drainage arrays are True
	wetlands_array = np.where((cti_binary_array == 1) & 
							  (lulc_array == 1) & 
							  (drainage_array == 1), 
							  1, 0)

	cti_binary_array, lulc_array, drainage_array = None, None, None
	wetlands_array = np.where(np.isnan(cti_array), np.nan, wetlands_array)

	return wetlands_array


def subset_wetlands(wetlands_array, gt, wetland_area_threshold):
	"""Subset wetland pixels based on threshold wetland area. The resulting 
        raster will contain of a subset of the potential wetlands originally 
        identified that have a greater area than the wetland_area_threshold. 
        Code adapted from: 
        gis.stackexchange.com/questions/29052/creating-a-raster-filter

    Parameters:
		wetlands_array (numpy array): a binary array identifying potential 
        	wetlands 
		gt (list): a list of geotransformation values
		wetland_area_threshold (integer): the area threshold for the minimum 
			size of a wetland

    Returns:
        None
    """

	print 'Subsetting wetlands'

	labelled_array, num_features = label(wetlands_array, structure=np.ones([3,3]))
	bin_count = np.bincount(labelled_array.flat)
	cellsize = gt[1] # assumes cell size is in meters
	subset_array = np.where(
	    bin_count[labelled_array] * cellsize ** 2 >= wetland_area_threshold, 1, 0)
	subset_array = np.where(np.isnan(wetlands_array), np.nan, wetlands_array)

	return subset_array

def wetlands_array2raster(wetlands_array, gt, proj, wetlands_uri):
	"""This function converts the array with cells identified as potential
	wetlands to a GeoTIFF raster.


	Parameters:
		wetlands_array (numpy array): a binary array identifying potential 
        	wetlands 
		gt (list): a list of geotransformation values
		proj (string): a projection string
		wetlands_uri (string): path to wetlands GeoTIFF raster

	Returns:
		None
	"""

	print 'Converting wetlands array to raster'

	util.array2raster(wetlands_array, gt, proj, wetlands_uri)

	return None

