import os
import gdal
import ogr
import numpy as np

def get_raster_projection(raster_uri):
    ds = gdal.Open(raster_uri)
    projection = ds.GetProjectionRef()

    return projection 


def raster2array(raster_uri):
    """Convert raster to array.

    Parameters:
        raster_uri (string): path to GeoTIFF

    Returns:
        array (numpy array): a numpy array of data contained within the first 
            band of the raster_uri
        gt (list): a dataset geotransform list
        projection (string): WKT describing the GeoTIFF projection
    """
    
    ds = gdal.Open(raster_uri)
    gt = ds.GetGeoTransform()
    band = ds.GetRasterBand(1)
    ndval = band.GetNoDataValue()
    array = band.ReadAsArray()
    array = np.where(array == ndval, np.nan, array)
    
    return array, gt


def array2raster(array, gt, projection, raster_uri):
    """Convert array to raster.

    Parameters:
        array (numpy array): a numpy array of data contained within the first 
            band of the raster_uri
        gt (list): a dataset geotransform list
        projection (string): WKT describing the GeoTIFF projection
        raster_uri (string): path to GeoTIFF

    Returns:
        None
    """

    # Create raster
    driver = gdal.GetDriverByName('GTiff')
    raster = driver.Create(raster_uri, array.shape[1], array.shape[0], 1, 
        gdal.GDT_Float32, options=['BIGTIFF=IF_SAFER'])
    raster.SetGeoTransform((gt[0], gt[1], 0, gt[3], 0, gt[5]))
    band = raster.GetRasterBand(1)
    
    # Write data to raster 
    array = np.where(np.isnan(array), 9999, array)
    band.WriteArray(array)
    band.SetNoDataValue(9999)
    band.ComputeStatistics(True)
    raster.SetProjection(projection)
    
    # Close and clean up dataset
    band.FlushCache()
    gdal.Dataset.__swig_destroy__(raster)
    band = raster = None

    return None


def shapefile2raster(
        shapefile_uri, gt, projection, output_directory, burn_field=None):
    """Convert shapefile to raster.

    Parameters: 
        shapefile_uri (string): path to shapefile
        gt (list): a dataset geotransform list
        projection (string): WKT describing the GeoTIFF projection
        output_directory (string): a directory to export the raster
        burn_field (string): the field in the shapefile to burn values into the
            raster. If left blank, uses Z values

    Returns:
        raster_uri: path to GeoTIFF
    """
    
    # Get shapefile data
    driver = ogr.GetDriverByName('ESRI Shapefile')
    ds = driver.Open(shapefile_uri)
    lyr = ds.GetLayer()
    srs = lyr.GetSpatialRef()
    x_min, x_max, y_min, y_max = lyr.GetExtent()
    x_res = int((x_max - x_min) / gt[1])
    y_res = int((y_max - y_min) / gt[1])
    
    # Create raster
    fn = os.path.basename(shapefile_uri).split('.')[0]
    raster_uri = os.path.join(output_directory, fn + '.tif')
    raster = gdal.GetDriverByName('GTiff').Create(
        raster_uri, x_res, y_res, 1, gdal.GDT_Float32, 
        options=['BIGTIFF=IF_SAFER'])
    raster.SetGeoTransform((x_min, gt[1], 0, y_max, 0, gt[5]))
    band = raster.GetRasterBand(1)
    band.SetNoDataValue(9999)
    raster.SetProjection(projection)
    if burn_field != None:
        gdal.RasterizeLayer(
            raster, [1], lyr, options=['ATTRIBUTE=%s' % burn_field])
    else:
        gdal.RasterizeLayer(
            raster, [1], lyr, burn_values=[1])
    band.ComputeStatistics(True)

    # Close and clean up dataset
    band.FlushCache()
    gdal.Dataset.__swig_destroy__(raster)
    band = raster = None

    return raster_uri

