import numpy as np
import pygeoprocessing as pygeo

import util
from paths import *

def execute(potential_wetlands_uri, actual_wetlands_uri):
	"""This function invokes a series of functions to validate the accuracy
	of the identified potential wetlands.

    Parameters:
		potential_wetlands_uri (string): path to potential wetlands GeoTIFF 
			raster
		actual_wetlands_uri (string): path to actual wetlands shapefile

    Returns:
       	error_dict (python dictionary): accuracy error dictionary
    """

	potential_wetlands_array, gt = util.raster2array(potential_wetlands_uri)
	actual_wetlands_array = shapefile2array(actual_wetlands_uri, gt, proj, 
		potential_wetlands_array.shape[0], potential_wetlands_array.shape[1])
	error_dict = calculate_error(potential_wetlands_array, actual_wetlands_array)
	print error_dict

	return error_dict


def shapefile2array(actual_wetlands_uri, gt, proj, nrows, ncols):
	"""This function converts the actual wetlands shapefile to a numpy
	array and aligns it to the potential wetlands GeoTIFF raster using the 
	geotransform, projection, and shape of the potential wetlands numpy array.

    Parameters:
		actual_wetlands_uri (string): path to actual wetlands shapefile
		gt (python list): a dataset geotransform list
		proj (string): WKT describing the GeoTIFF projection
		nrows (integer): number of rows in array
		ncols (integer): number of columns in array

    Returns:
       	actual_wetlands_array (numpy array): array for actual wetlands
    """

	# Convert actual_wetlands shapefile to raster
	output_directory = os.path.dirname(actual_wetlands_uri)
	actual_wetlands_uri = util.shapefile2raster(actual_wetlands_uri, gt, proj, 
												output_directory)

	# Resample actual_wetlands raster to align with potential_wetlands
	cellsize = gt[1]
	upper_left_x = gt[0]
	upper_left_y = gt[3]
	lower_right_x = upper_left_x + (cellsize * ncols)
	lower_right_y = upper_left_y - (cellsize * nrows)
	bbox = [upper_left_x, upper_left_y, lower_right_x, lower_right_y]
	actual_wetlands_resamp_uri = pygeo.temporary_filename(suffix='.tif')
	pygeo.resize_and_resample_dataset_uri(actual_wetlands_uri, bbox, cellsize, 
										  actual_wetlands_resamp_uri, 'nearest')

	# Convert resampled actual_wetlands raster to array
	actual_wetlands_array, gt = util.raster2array(actual_wetlands_resamp_uri)

	return actual_wetlands_array


def calculate_error(potential_wetlands_array, actual_wetlands_array):
	"""

    Parameters:
		potential_wetlands_array (numpy array): array for potential wetlands
		actual_wetlands_array (numpy array): array for actual wetlands

    Returns:
       	error_dict (python dictionary): accuracy error dictionary
    """
    		
	actual_wetlands_array = np.where((~np.isnan(potential_wetlands_array)) & 
									 (np.isnan(actual_wetlands_array)), 
									 0, actual_wetlands_array)
	
	def compare_arrays(poten_val, actual_val):
	   	potential_true = np.where(potential_wetlands_array==poten_val, 1, 0)
	   	actual_true = np.where(actual_wetlands_array==actual_val, 1, 0)
	   	pixel_count = np.nansum(np.multiply(potential_true, actual_true))

	   	return pixel_count

	error_dict = {}
	error_dict['Wetland_Wetland'] = compare_arrays(1, 1)
	error_dict['Wetland_NonWetland'] = compare_arrays(1, 0)
	error_dict['NonWetland_Wetland'] = compare_arrays(0, 1)
	error_dict['NonWetland_NonWetland'] = compare_arrays(0, 0)

	return error_dict 

