import os
import gdal
import numpy as np

from sklearn.ensemble import RandomForestRegressor

from paths import *

def main():
	### Working directory
	stem = os.path.join(landvalue_dir, 'Variables/')

	### Input dataset URIs
	input_uris_30m_dict = {
		'e911_count': os.path.join(stem, 'E911_SiteLocations_Count_30m.tif'),
		'road_density': os.path.join(stem, 'VT_Road_Centerline_Density_30m.tif'),
		'median_income': os.path.join(stem, 'ACS_2015_5YR_BG_50_VERMONT_B19013e1_30m.tif'),
		'urban_binary': os.path.join(stem, 'urban_binary_30m.tif'),
		'dist2urban': os.path.join(stem, 'distance2urban_30m.tif'),
		'lulc': os.path.join(stem, 'NCLD_2011_Reclassified_Vermont.tif'),
		'x_coords': os.path.join(stem, 'x_coords_30m.tif'),
		'y_coords': os.path.join(stem, 'y_coords_30m.tif'),

		'land_value': os.path.join(stem, 'landvalue_per_ha_observed_30m.tif')
		}
	input_uris_90m_dict = {
		'e911_count': os.path.join(stem, 'E911_SiteLocations_Count_90m.tif'),
		'road_density': os.path.join(stem, 'VT_Road_Centerline_Density_90m.tif'),
		'median_income': os.path.join(stem, 'ACS_2015_5YR_BG_50_VERMONT_B19013e1_90m.tif'),
		'urban_binary': os.path.join(stem, 'urban_binary_90m.tif'),
		'dist2urban': os.path.join(stem, 'distance2urban_90m.tif'),
		'lulc': os.path.join(stem, 'NLCD_2011_Reclassified_Vermont_90mResampled.tif'),
		'x_coords': os.path.join(stem, 'x_coords_90m.tif'),
		'y_coords': os.path.join(stem, 'y_coords_90m.tif'),

		'land_value': os.path.join(stem, 'landvalue_per_ha_observed_90m.tif')
		}

	### Output dataset URIs
	landvalue_30m_uri = os.path.join(stem[:-10], 'landvalue_predicted_30m.tif')
	landvalue_90m_uri = os.path.join(stem[:-10], 'landvalue_predicted_90m.tif')

	### Execute
	ntrees = 100
	execute(input_uris_90m_dict, landvalue_90m_uri, ntrees)

	return None


def execute(input_datasets_dict, output_uri, ntrees):
	array_dict, gt, proj, shape = read_data(input_datasets_dict)
	array_dict_full, array_dict_subset = clean_data(array_dict)
	X_full, X_subset = get_Xdata(array_dict_full, array_dict_subset)
	y_obs_full, y_obs_subset = get_Ydata(array_dict_full, array_dict_subset)
	# model = fit_model(X_subset, y_obs_subset, ntrees)
	# calculate_rsquared(model, y_obs_subset, X_subset)
	# y_pred_full = predict(model, X_full, array_dict_full)
	# export_raster(y_pred_full, gt, proj, shape, output_uri)

	return None


def read_data(input_datasets_dict):
	print 'Reading data'

	array_dict = {}
	for dataset in input_datasets_dict:
		raster_uri = input_datasets_dict[dataset]
		ds = gdal.Open(raster_uri)
		proj = ds.GetProjection()
		gt = ds.GetGeoTransform()
		band = ds.GetRasterBand(1)
		ndval = band.GetNoDataValue()
		array = band.ReadAsArray()
		array = np.where(array==ndval, np.nan, array)
		array_dict[dataset] = array 

	return array_dict, gt, proj, array.shape


def clean_data(array_dict):
	print 'Cleaning data'

	landvalue_array = array_dict['land_value']
	array_dict_full = {}
	array_dict_subset = {}
	
	for dataset in array_dict:
		array = array_dict[dataset]
		
		if dataset in ['e911_count', 'urban_binary']:
			array = np.where(np.isnan(array), 0, array)
			
		full_array = np.where(np.isnan(array), 0 , array)

		subset_array = array[~np.isnan(landvalue_array)]
		subset_array = np.where(np.isnan(subset_array), 0 , subset_array)

		array_dict_full[dataset] = full_array.flatten()
		array_dict_subset[dataset] = subset_array
		

	return array_dict_full, array_dict_subset


def get_Xdata(array_dict_full, array_dict_subset):
	print 'Getting X data'

	### Define X data
	x_list = ['e911_count', 'road_density', 'median_income', 'urban_binary', 
		 	  'dist2urban', 'lulc', 'x_coords', 'y_coords']
	X_full = []
	X_subset = []
	for x_var in x_list:
		data_full = array_dict_full[x_var]
		data_subset = array_dict_subset[x_var]

		### Print summary stats
		print x_var, ' - SUMMARY STATS'
		print '		Min: '    , np.min(data_subset)
		print '		Mean: '   , np.mean(data_subset)  
		print '		Median: ' , np.median(data_subset)
		print '		Max: '    , np.max(data_subset)
		print '		Std: '    , np.std(data_subset)
		print

		### Append data to lists
		X_full.append(data_full)
		X_subset.append(data_subset)

	### Swap axes
	X_full = np.swapaxes(np.array(X_full), 0, 1)
	X_subset = np.swapaxes(np.array(X_subset), 0, 1)

	return X_full, X_subset

def get_Ydata(array_dict_full, array_dict_subset):
	print 'Getting Y data'

	### Define y data
	y_obs_full = array_dict_full['land_value']
	y_obs_subset = array_dict_subset['land_value']

	### Print summary stats
	print 'land value - SUMMARY STATS'
	print '		Min: '    , np.min(y_obs_subset)
	print '		Mean: '   , np.mean(y_obs_subset)  
	print '		Median: ' , np.median(y_obs_subset)
	print '		Max: '    , np.max(y_obs_subset)
	print '		Std: '    , np.std(y_obs_subset)
	print

	return y_obs_full, y_obs_subset


def fit_model(X_subset, y_obs_subset, ntrees):
	print 'Fitting model'

	### Fit model
	model = RandomForestRegressor(n_estimators=ntrees, verbose=3)
	model.fit(X_subset, y_obs_subset)

	return model


def calculate_rsquared(model, y_obs_subset, X_subset):

	print 'Calculating R-squared'

	r_squared = model.score(X_subset, y_obs_subset)

	print '		R-squared: %f' %r_squared

	return None


def predict(model, X_full, array_dict_full):
	print 'Predicting values'

	### Predict
	y_pred_full = model.predict(X_full)

	### Change y_pred_full values where lulc==0 to nan
	lulc = array_dict_full['lulc']
	y_pred_full = np.where(lulc==0, np.nan, y_pred_full)

	return y_pred_full


def export_raster(y_pred_full, gt, projection, shape, output_uri):
	print 'Exporting raster'

	### Reshape predicted array
	y_pred_full = np.reshape(y_pred_full, shape)

	### Create raster
	driver = gdal.GetDriverByName('GTiff')
	rows = y_pred_full.shape[1]
	cols = y_pred_full.shape[0]
	raster = driver.Create(output_uri, rows, cols, 1, 
	    gdal.GDT_Float32, options=['BIGTIFF=IF_SAFER'])
	raster.SetGeoTransform((gt[0], gt[1], 0, gt[3], 0, gt[5]))
	band = raster.GetRasterBand(1)

	### Write data to raster 
	y_pred_full = np.where(np.isnan(y_pred_full), 9999, y_pred_full)
	band.WriteArray(y_pred_full)
	band.SetNoDataValue(9999)
	band.ComputeStatistics(True)
	raster.SetProjection(projection)

	### Close and clean up dataset
	band.FlushCache()
	gdal.Dataset.__swig_destroy__(raster)
	band = raster = None

	return None


if __name__ == '__main__':
	main()
