import os
import ogr
import time
import shutil
import pywinauto
import numpy as np
import geopandas as gpd

from datetime import datetime

import constants, ras_preprocessing, ras_controller, utils

from paths import *

def execute(args):

	### Pre-process data
	ras_controller.close_rasexe()
	args = preprocess_data(args)

	### Evaluate baseline model
	print '\nEvaluating baseline model'
	scenario_baseline = [0] * 199
	base_df = evaluate_intervention(scenario_baseline, args)
	uri = os.path.join(intervention_logbooks_dir, 'baseline.csv')
	base_df.to_csv(uri, index=False)
	args['base_df'] = base_df

	### Iterate through reaches
	for xs in range(199):

		### Iterate through interventions
		for i in range(3):
			print '\nEvaluating reach %s, intervention %s' % (str(xs), str(i+1))
			scenario_modified = scenario_baseline[:]
			scenario_modified[xs] = i + 1
			outputs_df = evaluate_intervention(scenario_modified, args)
			fn = '%s_%s.csv' %(str(xs), str(i+1))
			uri = os.path.join(intervention_logbooks_dir, fn)
			outputs_df.to_csv(uri, index=False)

	return None


def preprocess_data(args):	
	"""Pre-process data.

    Parameters:
        args (dict): dictionary containing GA parameters

    Returns:
        args (dict): dictionary containing GA parameters
    """

	print 'Pre-processing:'

	### Define inputs specific to ws
	ws = 'lewiscreek'
	ras_ws_dir = os.path.join(ras_dir, ws)
	ras_ws_base_dir = os.path.join(ras_ws_dir, 'basemodel')
	ras_ws_output_dir = os.path.join(ras_ws_dir, 'output')
	args['ras_proj_dir'] = ras_ws_output_dir 
	xs_uri = os.path.join(xs_dir, '%s_xs_indices.csv' %ws)
	pf_uri = os.path.join(pf_dir, 'peakflow_%s.csv' %ws)

	### Execute pre-propocessing functions
	base_mn_df = ras_preprocessing.preprocess_manningsn(ras_ws_base_dir)	
	e911_df = ras_preprocessing.preprocess_e911(e911_csv_uri, ras_ws_base_dir)	
	gl_df = ras_preprocessing.preprocess_grandlist(grandlist_uri)
	cla_df = ras_preprocessing.preprocess_cla(cla_uri)	
	e911_df = ras_preprocessing.join_e911_grandlist(e911_df, gl_df)
	e911_df = ras_preprocessing.standardize_propertyvalues(e911_df, cla_df)
	xs_df = ras_preprocessing.preprocess_crosssection_indices(xs_uri)
	pf_df = ras_preprocessing.preprocess_peakflow(pf_uri)

	print '		DEM rasters'
	baseline_dem_proj = utils.get_raster_projection(baseline_dem_uri)
	baseline_dem_arr, baseline_dem_gt = utils.raster2array(baseline_dem_uri)
	cut_dem_arr, cut_dem_gt = utils.raster2array(cut_dem_uri)
	cut_dem_arr = utils.align_arrays(cut_dem_arr, cut_dem_gt, 
									 baseline_dem_arr, baseline_dem_gt)
	xs_arr, xs_gt = utils.raster2array(xs_raster_uri)
	xs_arr = utils.align_arrays(xs_arr, xs_gt, baseline_dem_arr, 
							    baseline_dem_gt)

	### Put dataframes in data dictionary
	data_dict = {
		'base_mn_df': base_mn_df,
		'e911_df': e911_df,
		'xs_df': xs_df,
		'pf_df': pf_df,
		'baseline_dem_array': baseline_dem_arr,
		'cut_dem_array': cut_dem_arr,
		'xs_array': xs_arr,
		'terrain_gt': baseline_dem_gt,
		'terrain_proj': baseline_dem_proj
		}
	args['data_dict'] = data_dict

	return args


def evaluate_intervention(scenario, args):
	"""Genetic algorithm.

    Parameters:
    	scenario (list): encoding for individual restoration scenario
        args (dict): dictionary containing objective function parameters

    Returns:
		fitnesses (list): list of objectives' fitnesses
    """

	### Print start time
	counter = print_time()
	
	### Initialize variables in args
	ras_proj_dir = args['ras_proj_dir']
	data_dict = args['data_dict']
	if 'base_df' in args:
		base_df = args['base_df']

	### Initialize variables in data_dict
	e911_df = data_dict['e911_df']
	xs_df = data_dict['xs_df']
	pf_df = data_dict['pf_df']
	base_mn_df = data_dict['base_mn_df']
	baseline_dem_array = data_dict['baseline_dem_array']
	cut_dem_array = data_dict['cut_dem_array']
	xs_array = data_dict['xs_array']
	terrain_gt = data_dict['terrain_gt']
	terrain_proj = data_dict['terrain_proj']

	### Create outputs_df
	if 'base_df' in args:
		output_df = base_df.copy()
		output_df = output_df.rename(columns={
			'DEPTH_FT_2YR': 'DEPTH_FT_2YR_BASE', 
			'DEPTH_FT_10YR': 'DEPTH_FT_10YR_BASE', 
			'DEPTH_FT_25YR': 'DEPTH_FT_25YR_BASE', 
			'DEPTH_FT_100YR': 'DEPTH_FT_100YR_BASE'})
	else:
		output_df = e911_df.copy()
		fields = ['OBJECTID', 'TOWN', 'SITETYPE', 'PROPERTY_VALUE']
		output_df = output_df[fields]

	### If HEC-RAS is already open, close it
	ras_controller.close_rasexe()

	### Copy baseline HEC-RAS files to output directory
	ws = 'lewiscreek'
	ras_ws_dir = os.path.join(ras_dir, ws)
	ras_ws_base_dir = os.path.join(ras_ws_dir, 'basemodel')
	ras_ws_output_dir = os.path.join(ras_ws_dir, 'output')
	if os.path.exists(ras_ws_output_dir):
		shutil.rmtree(ras_ws_output_dir)
	shutil.copytree(ras_ws_base_dir, ras_ws_output_dir)	

	### Prepare HEC-RAS
	ras_controller.open_rasexe(ras_exe_uri)
	rc = ras_controller.open_rasproject(ras_ws_output_dir)

	### Edit Manning's N and Terrain
	trying = True
	loops = 0
	while trying:
		loops += 1 
		if loops > 3: 
				print 'CAUGHT IN WHILE LOOP - RESTARTING RAS'
				ras_controller.close_rasexe()
				ras_controller.open_rasexe(ras_exe_uri)
				rc = ras_controller.open_rasproject(ras_ws_output_dir)
		try:
			ras_window = edit_manningsn(base_mn_df, scenario, xs_df)
			trying = False
		except:
			pass

	# ### If the value 3 is in the scenario, edit the terrain raster...
	# ### if 3 in scenario:
		# edit_terrain(scenario, xs_df, baseline_dem_array, cut_dem_array, 
					 # xs_array, terrain_gt, terrain_proj, ras_window)

	### Edit the terrain raster for all scenarios
	edit_terrain(scenario, xs_df, baseline_dem_array, cut_dem_array, 
		 xs_array, terrain_gt, terrain_proj, ras_window)

	### Fix errors in cross-sections
	print 'Filtering cross-section points'
	ras_controller.filter_xspoints(ras_window)

	### Iterate through multiple flood events
	print 'Simulating flood events'
	flood_events = [2, 10, 25, 100]
	for flood_event in flood_events:
		print '		%s-year' %int(flood_event)
		alter_flowdata(ras_window, flood_event, pf_df)
		compute_steadyflow(ras_window, counter)
		output_df = get_depth(ras_proj_dir, counter, output_df, flood_event)

	### Calculate change from baseline
	if 'base_df' in args:
		output_df['DEPTH_FT_2YR_DELTA'] = (
			output_df['DEPTH_FT_2YR_BASE'] - output_df['DEPTH_FT_2YR'])
		output_df['DEPTH_FT_10YR_DELTA'] = (
			output_df['DEPTH_FT_10YR_BASE'] - output_df['DEPTH_FT_10YR'])
		output_df['DEPTH_FT_25YR_DELTA'] = (
			output_df['DEPTH_FT_25YR_BASE'] - output_df['DEPTH_FT_25YR'])
		output_df['DEPTH_FT_100YR_DELTA'] = (
			output_df['DEPTH_FT_100YR_BASE'] - output_df['DEPTH_FT_100YR'])

	return output_df	


def print_time():
	"""Print the time at which the objective function was started.

    Parameters:
        None

    Returns:
        counter (int): number of function evaluations
    """

	### Print start time
	time = datetime.now().strftime('%H:%M:%S')
	time = datetime.strptime(time, '%H:%M:%S').strftime("%I:%M:%S %p")
	print 'Starting scenario %s' %(time)

	return None


def edit_manningsn(base_mn_df, scenario, xs_df):
	"""Makes call to ras_controller to edit Manning's N table in HEC-RAS to 
		reflect individual scenario.

	Parameters:
	    base_mn_df (pandas dataframe): baseline Manning's N table
	    scenario (list): encoding for individual restoration scenario

	Returns:
	    ras_window (object): HEC-RAS window object
	"""

	### Print 
	print "Editing Manning's N table"

	### Copy tables to new dataframes
	scenario_mn_df = base_mn_df.copy()

	### Initialize Manning's N values for specif land cover classes
	woody_wetlands_mn = constants.mannings_n_lulc['woody_wetlands']
	herb_wetlands_mn = constants.mannings_n_lulc['herbaceous_wetlands']
	mixed_mn = constants.mannings_n_lulc['mixed_forest']
	stream_mn = constants.mannings_n_lulc['stream']
	deciduous_mn = constants.mannings_n_lulc['deciduous_forest']
	evergreen_mn = constants.mannings_n_lulc['evergreen_forest']

	### Edit Manning's N table in dataframe
	intervention_code = max(scenario)
	scenario_index = scenario.index(intervention_code)
	base_mn = base_mn_df.loc[[scenario_index]]

	### Replace Manning's N coefficients in base_mn_df	
	if intervention_code == 1:
		### Replace all nodes in XS with deciduous Manning's N value
		scenario_mn_df[base_mn.notnull()] = deciduous_mn
		
		### Replace stream nodes in XS with original stream value
		scenario_mn_df[base_mn == stream_mn] = stream_mn

		### Replace forested nodes in XS with original forest value
		scenario_mn_df[base_mn == mixed_mn] = mixed_mn
		scenario_mn_df[base_mn == evergreen_mn] = evergreen_mn

		### Replace forested nodes in XS with original wetland value
		scenario_mn_df[base_mn == woody_wetlands_mn] = woody_wetlands_mn
		scenario_mn_df[base_mn == herb_wetlands_mn] = herb_wetlands_mn

	if intervention_code in [2, 3]:
		### Replace all nodes in XS with woody wetlands value
		scenario_mn_df[base_mn.notnull()] = woody_wetlands_mn
		
		### Replace stream nodes in XS with original stream value
		scenario_mn_df[base_mn == stream_mn] = stream_mn

		### Replace forested nodes in XS with original forest value
		scenario_mn_df[base_mn == deciduous_mn] = deciduous_mn
		scenario_mn_df[base_mn == mixed_mn] = mixed_mn
		scenario_mn_df[base_mn == evergreen_mn] = evergreen_mn

		### Replace wetlands nodes in XS with original wetland value
		scenario_mn_df[base_mn == herb_wetlands_mn] = herb_wetlands_mn


	### Edit Manning's N table in HEC-RAS 
	ras_window = ras_controller.edit_manningsn(scenario_mn_df)

	return ras_window


def edit_terrain(scenario, xs_df, baseline_dem_array, 
				 cut_dem_array, xs_array, gt, proj, ras_window):
	"""Creates new terrain layer for scenario and make calls to ras_controller
		to upload new terrain layer into HEC-RAS.

	Parameters:
		scenario (list): encoding for individual restoration scenario
		xs_df (pandas dataframe): cross-section dataframe
		baseline_dem_array (numpy array):
		cut_dem_array(numpy array): array indicating elevation when all HEC-RAS 
			cross-sections are cut
		xs_array (numpy array): array indicating the river station IDs for 
			HEC-RAS cross-sections
		gt (list): DEM GDAL geotransform list
		proj (string): SRS projection for DEM
		ras_window (object): HEC-RAS window object

	Returns:
		None
	"""

	### Copy baseline_dem_array to dem_array
	print 'Editing terrain'
	dem_array = baseline_dem_array.copy()

	### Edit terrain array
	if 3 in scenario:	
		scenario_index = scenario.index(3)
		rs = float(xs_df['STATION'][xs_df['INDEX']==scenario_index])
		dem_array = np.where(xs_array==rs, cut_dem_array, dem_array)
	
	### Export edited dem array to raster
	fn = 'edited_dem.tif'
	edited_dem_uri = os.path.join(ras_dir, 'lewiscreek/output', fn)
	utils.array2raster(dem_array, gt, proj, edited_dem_uri)

	### Open edited dem raster in HEC-RAS
	print 'Adding edited terrain to HEC-RAS'
	ras_controller.open_editedterrain(ras_window, edited_dem_uri)

	### Adjust cross-section data to edited dem raster
	print 'Editing cut profiles'
	geom_window = ras_controller.edit_cutprofiles(ras_window)

	return None


def alter_flowdata(ras_window, flood_event, pf_df):
	"""Change flow data in HEC_RAS to specified flood event.

	Parameters:
		ras_window (object): HEC-RAS window object
		flood_event (int): frequency of flood event occurance

	Returns:
		None
	"""

	### Subset peak flow dataframe to just series for particular flood event
	pf_series = pf_df['%sYR' %str(flood_event)]

	### Alter flow steady-flow data
	ras_controller.change_steadyflow(ras_window, flood_event, pf_series)

	return None


def compute_steadyflow(ras_window, counter):
	"""Makes call to ras_controller to compute steady-flow scenario in HEC-RAS.

    Parameters:
        ras_window (object): HEC-RAS window object
        counter (int): number of function evaluations

    Returns:
        None
    """

	### Compute steady-flow analysis
	ras_controller.compute_steadyflow(ras_window, counter)
	
	return None


def get_depth(ras_proj_dir, counter, output_df, flood_event):
	"""Makes call to ras_controller to compute steady-flow scenario in HEC-RAS.

    Parameters:
        ras_proj_dir (string): HEC-RAS project directory
        counter (int): number of function evaluations
        output_df (pandas dataframe): depth table outputs

    Returns:
        output_df (pandas dataframe): depth table outputs
    """

	### Get depth outputs raster URI
	output_dir = os.path.join(ras_proj_dir, str(counter))
	output_raster_fn = 'Depth (PF 1).lcfb_hflidar.tif'
	output_raster_uri = os.path.join(output_dir, output_raster_fn)

	### Open depth outputs raster to array
	depth_array, depth_gt = utils.raster2array(output_raster_uri)

	### Define pandas coordinate functions
	def x_coord(point):
		return point.x

	def y_coord(point):
		return point.y

	### Extract depth to points
	e911_df = gpd.read_file(e911_shp_uri)
	x_coords = np.array(e911_df['geometry'].apply(x_coord))
	y_coords = np.array(e911_df['geometry'].apply(y_coord))
	col_indices = np.floor((x_coords - depth_gt[0]) / depth_gt[1]).astype(int)
	row_indices = np.floor((depth_gt[3] - y_coords) / depth_gt[1]).astype(int)
	depth_m = depth_array[row_indices, col_indices]
	depth_ft = depth_m * constants.meters2ft
	depth_ft = np.nan_to_num(depth_ft)

	### Clean up
	depth_column = 'DEPTH_FT_%sYR' %str(flood_event)
	e911_df[depth_column] = depth_ft
	e911_df = e911_df[['OBJECTID', depth_column]]
	output_df = output_df.merge(e911_df, on='OBJECTID', how='right')

	### Delete ouputs directory
   	shutil.rmtree(output_dir) 

	return output_df


if __name__ == '__main__':
	main()
	
	

