import json
import numpy as np
import pandas as pd

import constants, utils

from paths import *

def process_rasoutputs():
	print 'Post-processing RAS data:'

	### Create dictionary in which to append data
	base_dict = {}
	data_dict = {'XS':[], 'INTERVENTION':[], 'DAMAGECOST_EW-2':[],
		'DAMAGECOST_EW-1':[], 'DAMAGECOST_EW0':[], 'DAMAGECOST_EW1':[], 
		'DAMAGECOST_EW2':[], 'RESTORATION_COST':[], 'PV_ADJUSTED':[]}
	
	### Read in easement costs data from csv
	ec_df = pd.read_excel(easementcosts_uri)

	### Read in depth-damage functions from CSV & 
	### create columns for slope values
	ddf_df = pd.read_csv(depthdamage_funcs_uri)
	dd_funcs = list(ddf_df.columns)
	dd_funcs.remove('DEPTH_FT')
	for func in dd_funcs:
		if func != 'DEPTH_FT':
			ddf_df[func + '_slope'] = ddf_df[func].shift(-1) - ddf_df[func]

	### Get list of intervention CSVs
	intervention_csvs = os.listdir(intervention_logbooks_dir)

	### Get logbooks fn list and move 'baseline.csv' to beginning
	logbooks_list = os.listdir(intervention_logbooks_dir)
	logbooks_list.insert(0, logbooks_list.pop(
		logbooks_list.index('baseline.csv')))

	### Iterate through intervention CSVs and evaluate
	for i, fn in enumerate(logbooks_list):
		print '		' + fn
		if 'baseline' not in fn:
			xs_intervention = fn.split('.')[0] # Remove '.csv'
			xs = int(xs_intervention.split('_')[0]) # Get cross-section
			intervention = int(xs_intervention.split('_')[1]) # Get intervention
			data_dict['XS'].append(xs)
			data_dict['INTERVENTION'].append(intervention) 

		### Read intervention csv to Pandas dataframe
		df = csv2df(fn)

		### Iterate through flood events
		flood_events = [2, 10, 25, 100]
		for event in flood_events:
			### Calculate percent damage to properties based on inundation depth
			df = depth2damage(df, ddf_df, event)

			### Calculate monetary damage to properties based on percent damage
			df = percdamage2dollardamage(df, event)

		### Calculate expected annual damages using numerical integration of 
		### monetary damages from flood events and annual probability
		df = trapezoidal_integration(df)
		
		### Calculate present value of expected annual damages over 100 year
		### time horizon and 7% discount rate
		df = present_value(df)	
		if 'baseline' not in fn:
			data_dict['PV_ADJUSTED'].append(np.nansum(df['PV_ADJUSTED']))
		else:
			base_dict['PV_ADJUSTED'] = np.nansum(df['PV_ADJUSTED']) 
		
		### Iterate through equity weights and estimate equity-weighted 
		### damage costs for each property
		for n in [-2, -1, 0, 1, 2]:
			damagecosts_ew = equityweighted_damagecosts(df, n)
			df.to_csv(os.path.join(intervention_logbooks_dir, fn), index=False)

			df['DAMAGECOST_EW%s' %(str(n))] = damagecosts_ew
			total_damagecost_ew = np.nansum(damagecosts_ew)	
			if 'baseline' not in fn:
				data_dict['DAMAGECOST_EW%s' %(str(n))].append(total_damagecost_ew)
			else:
				base_dict['DAMAGECOST_EW%s' %(str(n))] = total_damagecost_ew

		### Evaluate restoration costs for intervention and append to data_dict
		if 'baseline' not in fn:
			total_cost = evaluate_cost(ec_df, xs, intervention)
			data_dict['RESTORATION_COST'].append(total_cost)

		### Save base_dict to JSON
		else: 	
			with open(base_json_uri, 'w') as json_file:
				json.dump(base_dict, json_file, 
					default=utils.obj2json, indent=2)

	### Cleanup dataframe and save to csv
	dict2csv(data_dict, base_dict)
	
	return None


def csv2df(fn):
	### Open CSV and read to dataframe
	csv_uri = os.path.join(intervention_logbooks_dir, fn)
	df = pd.read_csv(csv_uri)
	prop_value = df['PROPERTY_VALUE']
	df = df[~prop_value.isnull()]

	return df


def depth2damage(df, ddf_df, flood_event):
    """Estimate percent damages for properties based on flood depth.

    Parameters:
        df (pandas dataframe): depth table outputs
        ddf_df (pandas dataframe): depth-damage functions

    Returns:
        df (pandas dataframe): outputs table
    """

    ### Create percent damage column in output df
    percent_damage_str = 'DAMAGE_PERCENT_%sYR' %str(flood_event)
    df[percent_damage_str] = np.nan

    ### Loop through properties 
    for i, row in df.iterrows():
        depth_ft = row['DEPTH_FT_%sYR' %str(flood_event)]
        site_type = row['SITETYPE']
        
        ### If site type is accessory building, then percent damage is 0
        if site_type in constants.sitetype2damagefunc_dict.keys():
            ### Choice damage function by on type of site
            func = constants.sitetype2damagefunc_dict[site_type]
            
            ### Estimate damage based on step-wise linear function
            b0 = ddf_df[func][ddf_df['DEPTH_FT']==int(depth_ft)]
            b1 = ddf_df[func + '_slope'][ddf_df['DEPTH_FT']==int(depth_ft)]
            
            if depth_ft > 0:
                percent_damage = b0 + b1 * (depth_ft % 1)
                # df.set_value(i, percent_damage_str, percent_damage / 100)
                df.at[i, percent_damage_str] = percent_damage / 100
            else:
                # df.set_value(i, percent_damage_str, 0)
                df.at[i, percent_damage_str] = 0

        else:
            # df.set_value(i, percent_damage_str, 0)
    		df.at[i, percent_damage_str] = 0

    return df    


def percdamage2dollardamage(df, event):
	property_value = df['PROPERTY_VALUE']
	percent_damage = df['DAMAGE_PERCENT_%sYR' %str(event)]
	dollar_damage = property_value * percent_damage
	df['DOLLAR_DAMAGE_%sYR' %str(event)] = dollar_damage

	return df


def trapezoidal_integration(df):
	### Calculated expected annual damages (ead) using numerical integration
	### See Olsen et al. 2015 - Method B, Trapezoidal Rule
	
	p = [0.01, 0.04, 0.1, 0.5]
	D = [df['DOLLAR_DAMAGE_2YR'],  df['DOLLAR_DAMAGE_10YR'], 
		 df['DOLLAR_DAMAGE_25YR'], df['DOLLAR_DAMAGE_100YR']]
	
	ead = 0 
	for j in range(3):
		ead += (p[j+1] - p[j]) * (D[j+1] + D[j]) * 0.5

	df['EAD'] = ead

	return df


def present_value(df, time_horizon=100, discount_rate=0.07):
    """Calculate present value of avoided damage costs over a paricular 
        discounted time horizon.

    Parameters:
        ead (float): expected annual damages
        time_horizon (int): number of years to calculate present value over
        discount_rate (float): discount rate 
    
    Returns:
        present_value: present value of avoided damages over time horizon
    """

    pv = 0
    for t in range(time_horizon):
        pv += (df['EAD'] * ((1+discount_rate)**(-1*t)))

    df['PV'] = pv

    ### Adjust PV if PV is greater than value of property
    df['PV_ADJUSTED'] = np.where(df['PV'] > df['PROPERTY_VALUE'],
    							 df['PROPERTY_VALUE'], df['PV'])

    return df


def equityweighted_damagecosts(df, n):
    """Calculate loss of equity-weighted (ew) utility

    Parameters:
        c0 (vector): a vector with 'initial' consumption values
        d (vector): a vector with 'damage' values
        n (float): aversion to inequality parameter

    Returns:
        u_rel (vector): a vector with relative utility values after damage
    """

    ### Convert n to type float
    n = float(n)

    ### Get "initial consumption values"
    c0 = df['PROPERTY_VALUE'].copy()
    if n != 0:
    	c0 *= 1000

    ### Get "damage" values
    d = df['PV'].copy()

    ### Calculate change in consumption from damage
    c1 = c0 - d

    ### If n=1, use log10 function
    if n == 1:
        u0 = np.log10(c0)
        u1 = np.log10(c1)

    else:
        u0 = np.power(c0, (1 - n)) / (1 - n)
        u1 = np.power(c1, (1 - n)) / (1 - n)

    ### Calculate relative change in utility
    u_rel = u1 - u0
    
    ### Convert 'change' in utility to 'lost' utility
    u_rel *= -1

    return u_rel


def evaluate_cost(ec_df, xs, intervention):
	
	### Get easement cost data for cross-section
	ec_xs_df = ec_df[ec_df['ORIG_FID']==xs]
	easement_cost = float(ec_xs_df['COST_TOTAL'])

	### Get intervention cost
	if intervention == 1: # Forest revegetation
		implementation_cost = constants.forestreveg_cost_dollars_ha
	
	if intervention == 2: # Wetland revegetation
		implementation_cost = constants.wetlandreveg_cost_dollars_ha
	
	if intervention == 3: # Floodplain lowering 
		implementation_cost = constants.floodplainlowering_cost_dollars_ha
	
	### Scale implementation cost by area of cross-section
	implementation_cost *= float(ec_xs_df['HECTARES'])
		
	### Add easement cost and implementation cost
	total_cost = easement_cost + implementation_cost

	return total_cost


def dict2csv(data_dict, base_dict):
	### Convert data_dict into dataframe
	output_df = pd.DataFrame(data_dict)

	output_df['AVOIDEDCOST_ADJUSTED'] = (
		base_dict['PV_ADJUSTED'] - data_dict['PV_ADJUSTED'])

	### Add AVOIDEDCOST columns to output dataframe
	for n in [-2, -1, 0, 1, 2]:
		output_df['AVOIDEDCOST_EW%s' %str(n)] = (
			base_dict['DAMAGECOST_EW%s' %str(n)] - 
			data_dict['DAMAGECOST_EW%s' %str(n)] )
	
	### Sort rows and set order of columns in dataframe
	output_df = output_df.sort_values(['XS', 'INTERVENTION'])
	output_df = output_df[
		['XS', 'INTERVENTION', 
		'RESTORATION_COST', 'AVOIDEDCOST_ADJUSTED',
		'DAMAGECOST_EW-2',  'AVOIDEDCOST_EW-2',
		'DAMAGECOST_EW-1',  'AVOIDEDCOST_EW-1',
		'DAMAGECOST_EW0' ,  'AVOIDEDCOST_EW0' ,
		'DAMAGECOST_EW1' ,  'AVOIDEDCOST_EW1' ,
		'DAMAGECOST_EW2' ,  'AVOIDEDCOST_EW2' ,
		]]

	### Save dataframe to csv
	output_df.to_csv(ga_inputs_csv_uri, index=False)

	return None
