import os
import inspect
import numpy as np
import pandas as pd
import geopandas as gpd
import pygeoprocessing as pygeo

import constants,  utils
from paths import *

def preprocess_manningsn(ras_basemodel_dir):
	"""Preprocess Manning's N data.
	
	Parameters:
		ras_basemodel_dir (string): directory for HEC-RAS basemodel

	Returns:
		base_mn_df (pandas dataframe): Baseline Manning's N table
	"""

	print '		Mannings N data'

	### Define URI for baseline Manning's N csv
	manningsn_uri = os.path.join(ras_basemodel_dir, 'basemodel_manningsn.csv')	

	### Open Manning's N csv to dataframe
	base_mn_df = pd.read_csv(manningsn_uri)

	### Remove unneeded columns
	base_mn_df = base_mn_df.iloc[:,4:]

	return base_mn_df


def preprocess_e911(e911_csv_uri, ras_ws_base_dir, e911_shp_uri=e911_shp_uri):
	"""Preprocess E911 site-locations data.
	
	Parameters:
		e911_csv_uri (string): path to E911 site-locations csv
		ras_ws_base_dir (string): 

	Optional parameters:
		e911_shp_uri (string): directory for HEC-RAS basemodel

	Returns:
		e911_df (pandas dataframe): E911 table
	"""

	print '		E911 data'

	if os.path.exists(e911_csv_uri):
		e911_df = pd.read_csv(e911_csv_uri)

	else:
		### Open e911_uri to geopandas dataframe
		e911_df = gpd.read_file(e911_shp_uri)

		### Extract x and y coordinates from points
		e911_df['X'] = e911_df.geometry.map(lambda p: p.x)
		e911_df['Y'] = e911_df.geometry.map(lambda p: p.y)

		### Subset columns
		e911_df = e911_df[['OBJECTID', 'SITETYPE', 'X', 'Y', 'PRIMARYADD', # E911 columns
						   'PSPAN', 'PARCELNU_1', 'TOWN']] # Parcel columns

		### Change 'PARCELNU_1' column name to 'PARCELNUM'
		e911_df = e911_df.rename(columns = {'PARCELNU_1':'PARCELNUM'})

		### Remove dashes from PSPAN column 
		e911_df['PSPAN'] = e911_df['PSPAN'].str.replace('-', '')

		### Convert PSPAN column from object to numberic dtype
		e911_df['PSPAN'] = pd.to_numeric(e911_df['PSPAN'], errors='coerce')

		# Export to CSV
		e911_df.to_csv(e911_csv_uri, index=False)

	return e911_df


def preprocess_grandlist(grandlist_uri):
	"""Preprocess VT 2016 Grand List data.
	
	Parameters:
		grandlist_uri (string): path to Grand List csv

	Returns:
		gl_df (pandas dataframe): Grand List table
	"""

	print '		Grand List data'

	### Open grandlist_uri text file
	gl_df = pd.read_table(grandlist_uri, delimiter=',', low_memory=False)

	### Subset columns of interest
	gl_df = gl_df[['TNAME',    # Town name
				   'SPAN',     # School Property Account Number
				   'PARCID',   # Parcel ID
				   'REAL_FLV', # Listed Real Value (Full)
				   'ADDRGL1'   # Mailing Address 1
					]]

	return gl_df


def preprocess_cla(cla_uri):
	"""Preprocess VT town's Common Level of Appraisal data.
	
	Parameters:
		cla_uri (string): path to Common Level of Appraisal csv

	Returns:
		cla_df (pandas dataframe): Common Level of appraisal table
	"""

	print '		Common Level of Appraisal data'

	### Read common Common Level of Appraisal dataset
	cla_df = pd.read_csv(cla_uri)

	### Convert percentages to decimals
	cla_df['CLA_PERCENT'] = cla_df['CLA_PERCENT'] / 100

	return cla_df


def preprocess_landvalue(lv_uri):
	"""Preprocess land value data.
	
	Parameters:
		lv_uri (string): path to land value csv

	Returns:
		lv_df (pandas dataframe): land value table
	"""

	print '		Land Value'
	
	### Open land value csv to pandas dataframe
	lv_df = pd.read_csv(lv_uri)
	lv_df['STATION'] = lv_df['River_Stat']
	lv_df = lv_df[['STATION', 'HECTARES', 'VALUE']]
	
	return lv_df

def preprocess_crosssection_indices(xs_uri):
	"""Preprocess cross-section indices data.
	
	Parameters:
		xs_uri (string): path to cross-section indices csv

	Returns:
		xs_df (pandas dataframe): cross-section indices table
	"""

	print '		Cross-Section Indices'

	### Open cross-section indices csv to pandas dataframe
	xs_df = pd.read_csv(xs_uri)

	return xs_df


def preprocess_peakflow(pf_uri):
	"""Preprocess peak flow data.
	
	Parameters:
		pk_uri (string): path to peak flow data csv

	Returns:
		pk_df (pandas dataframe): peak flow data table
	"""

	print '		Peak Flow Data'

	### Open cross-section indices csv to pandas dataframe
	pf_df = pd.read_csv(pf_uri, sep='\t')

	return pf_df


def join_e911_grandlist(e911_df, gl_df):
	"""Join E911 Site Locations data and Grand List data.
	
	Parameters:
		e911_df (pandas dataframe): E911 site-locations table
		gl_df (pandas dataframe): Grand List table

	Returns:
		e911_df (pandas dataframe): Joined E911 and Grand List table
	
	"""

	print '		Joining E911 and Grand List datasets'

	### Clean town names in e911_df
	e911_df.loc[e911_df['TOWN'] == 'ST ALBANS TOWN', 'TOWN'] = 'STALBANSTOWN'

	### List towns of interest
	town_list = ['BRISTOL', 
				 'CHARLOTTE', 
				 'FERRISBURGH',
				 'HIGHGATE',
				 'HINESBURG',
				 'HUNTINGTON',
				 'MONKTON',
				 'STARKSBORO',
				 'SWANTON',
				 'WARREN'
				 ]

	### Iterate through towns
	temp_df_list = []
	for town in town_list:
		### Copy original dataframe to temp dataframes
		e911_temp = e911_df.copy()
		gl_temp = gl_df.copy()

		### Subset to town 
		e911_temp = e911_temp[e911_temp['TOWN']==town]

		# Use town-specific join
		if town == 'BRISTOL':
			### Remove leading zero from PARCELNUM column in e911_df
			e911_temp['PARCELNUM'] = e911_temp['PARCELNUM'].map(lambda x: x[1:])

			### Manually merge e911_df and gl_temp, based on addresses
			e911_temp.loc[e911_temp['PRIMARYADD'] == '2360 HARDSCRABBLE RD', 'PARCELNUM'] = '20103000a'
			e911_temp.loc[e911_temp['PRIMARYADD'] == '415 UPPER MEEHAN RD',  'PARCELNUM'] = '40137000a'

			### Subset gl_temp to town of interest
			gl_temp = gl_temp[gl_temp['TNAME']=='Bristol']

			gl_temp.loc[gl_temp['ADDRGL1'] == '1026 MEEHAN ROAD',       'PARCID'] = '40137000' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '2308 HARDSCRABBLE ROAD', 'PARCID'] = '20104000' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '2352 HARDSCRABBLE RD',   'PARCID'] = '20103000' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '2360 HARDSCRABBLE ROAD', 'PARCID'] = '20103000a' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '2612 MONKTON ROAD',      'PARCID'] = '30119900'
			gl_temp.loc[gl_temp['ADDRGL1'] == '415 UPPER MEEHAN ROAD',  'PARCID'] = '40137000a' 

			### Use PARCELNUM and PARCID columns to merge
			e911_temp = e911_temp.merge(gl_temp, left_on='PARCELNUM', right_on='PARCID', how='left')
			
		if town == 'CHARLOTTE':
			### Use SPAN columns to join e911_df and gl_temp
			e911_temp = e911_temp.merge(gl_temp, left_on='PSPAN', right_on='SPAN', how='left')

		if town == 'FERRISBURGH':
			### Use SPAN columns to join e911_df and gl_temp
			e911_temp = e911_temp.merge(gl_temp, left_on='PSPAN', right_on='SPAN', how='left')

		if town == 'HIGHGATE':
			### Remove trailing dash from PARCELNUM column in e911_df
			gl_temp['PARCID'] = gl_temp['PARCID'].map(lambda x: x[:-1])

			### Use PARCELNUM and PARCID columns to merge
			e911_temp = e911_temp.merge(gl_temp, left_on='PARCELNUM', right_on='PARCID', how='left')

		if town == 'HINESBURG':
			### Use SPAN columns to join e911_df and gl_temp
			e911_temp = e911_temp.merge(gl_temp, left_on='PSPAN', right_on='SPAN', how='left')

		if town == 'HUNTINGTON':
			e911_temp.loc[e911_temp['PRIMARYADD'] == '49 DOGWOOD LN', 'PSPAN'] = 30309610467
			e911_temp.loc[e911_temp['PRIMARYADD'] == '64 DOGWOOD LN', 'PSPAN'] = 30309610225
			e911_temp.loc[e911_temp['PRIMARYADD'] == '67 DOGWOOD LN', 'PSPAN'] = 30309610561

			### Use SPAN columns to join e911_df and gl_temp
			e911_temp = e911_temp.merge(gl_temp, left_on='PSPAN', right_on='SPAN', how='left')

		if town == 'MONKTON':
			### Use SPAN columns to join e911_df and gl_temp
			e911_temp = e911_temp.merge(gl_temp, left_on='PSPAN', right_on='SPAN', how='left')

		if town == 'STARKSBORO':
			### Subset gl_temp to town of interest
			gl_temp = gl_temp[gl_temp['TNAME']=='Starksboro']

			### Manually merge e911_df and gl_temp, based on addresses
			gl_temp.loc[gl_temp['ADDRGL1'] == '106 BIRCH LANE',                    'ADDRGL1'] = '106 BIRCH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '11 CHERRY STREET',                  'ADDRGL1'] = '11 CHERRY LN'
			gl_temp.loc[gl_temp['ADDRGL1'] == '11 DOGWOOD LANE',                   'ADDRGL1'] = '11 DOGWOOD LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '117 ASH LANE',                      'ADDRGL1'] = '117 ASH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '117 BIRCH LANE',                    'ADDRGL1'] = '117 BIRCH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '140 BIRCH LANE',                    'ADDRGL1'] = '140 BIRCH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '147 ASH LANE',                      'ADDRGL1'] = '147 ASH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '151 BIRCH LANE',                    'ADDRGL1'] = '151 BIRCH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '153 LAZY BROOK DRIVE',              'ADDRGL1'] = '153 LAZY BROOK DR' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '171 LAZYBROOK  DRIVE',              'ADDRGL1'] = '171 LAZY BROOK DR' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '19 BIRCH LANE',                     'ADDRGL1'] = '19 BIRCH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '23 ASH LANE',                       'ADDRGL1'] = '23 ASH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '233 STATES PRISON HOLLOW ROAD EXT', 'ADDRGL1'] = '233 STATES PRISON HOLLOW EXT' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '24 CHERRY LANE',                    'ADDRGL1'] = '24 CHERRY LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '24 DOGWOOD LANE',                   'ADDRGL1'] = '24 DOGWOOD LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '25 DOGWOOD LANE',                   'ADDRGL1'] = '25 DOGWOOD LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '29 CHERRY LANE',                    'ADDRGL1'] = '29 CHERRY LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '291 STATES PRISON HOLLOW EXT.',     'ADDRGL1'] = '291 STATES PRISON HOLLOW EXT' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '3501 HINESBURG HOLLOW  ROAD',       'ADDRGL1'] = '3501 HINESBURG HOLLOW RD' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '37 DOGWOOD LANE',                   'ADDRGL1'] = '37 DOGWOOD LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '38 DOGWOOD LANE',                   'ADDRGL1'] = '38 DOGWOOD LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '4 BIRCH LANE',                      'ADDRGL1'] = '4 BIRCH LN'
			gl_temp.loc[gl_temp['ADDRGL1'] == '43 CHERRY LANE',                    'ADDRGL1'] = '43 CHERRY LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '44 BIRCH LANE',                     'ADDRGL1'] = '44 BIRCH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '45 ASH LANE',                       'ADDRGL1'] = '45 ASH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '48 DOGWOOD LANE',                   'ADDRGL1'] = '48 DOGWOOD LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '50 CHERRY LANE',                    'ADDRGL1'] = '50 CHERRY LN'
			gl_temp.loc[gl_temp['ADDRGL1'] == '55 CHERRY LANE',                    'ADDRGL1'] = '55 CHERRY LN'
			gl_temp.loc[gl_temp['ADDRGL1'] == '56 SHAMROCK DRIVE',                 'ADDRGL1'] = '56 SHAMROCK DR'
			gl_temp.loc[gl_temp['ADDRGL1'] == '62 CHERRY LANE',                    'ADDRGL1'] = '62 CHERRY LN'
			gl_temp.loc[gl_temp['ADDRGL1'] == '75 BIRCH LANE',                     'ADDRGL1'] = '75 BIRCH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '77 ASH LANE',                       'ADDRGL1'] = '77 ASH LN' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '8 CHERRY LANE',                     'ADDRGL1'] = '8 CHERRY LN'
			gl_temp.loc[gl_temp['ADDRGL1'] == '87 LAZY BROOK DRIVE',               'ADDRGL1'] = '87 LAZY BROOK DR'

			gl_temp.loc[gl_temp['ADDRGL1'] == '22 VARNEY HILL ROAD',           'ADDRGL1'] = '22 VARNEY HILL RD' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '261 ROBERT YOUNG ROAD',         'ADDRGL1'] = '261 ROBERT YOUNG RD' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '294 PARSONAGE ROAD',            'ADDRGL1'] = '294 PARSONAGE RD' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '306 STATES PRISON HOLLOW ROAD', 'ADDRGL1'] = '306 STATES PRISON HOLLOW RD'
			gl_temp.loc[gl_temp['ADDRGL1'] == '4696 VT ROUTE 116',             'ADDRGL1'] = '4696 VT ROUTE 116' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '57 CEMETERY ROAD',              'ADDRGL1'] = '57 CEMETERY RD' 
			gl_temp.loc[gl_temp['ADDRGL1'] == '75 VARNEY HILL ROAD',           'ADDRGL1'] = '75 VARNEY HILL RD' 

			### Use PRIMEARYADD and ADDRGL1 to merge
			e911_temp = e911_temp.merge(gl_temp, left_on='PRIMARYADD', right_on='ADDRGL1', how='left')
			
			### Impute missing data with Zillow data
			e911_temp.loc[e911_temp['PRIMARYADD']=='10 DOGWOOD LN', 'REAL_FLV'] = 21600
			e911_temp.loc[e911_temp['PRIMARYADD']=='274 STATES PRISON HOLLOW RD', 'REAL_FLV'] = 85922
			e911_temp.loc[e911_temp['PRIMARYADD']=='36 CHERRY LN', 'REAL_FLV'] = 27783
			e911_temp.loc[e911_temp['PRIMARYADD']=='75 IRELAND RD', 'REAL_FLV'] = 100672

		if town == 'SWANTON':
			### Manually merge e911_df and gl_temp, based on addresses
			gl_temp.loc[gl_temp['PARCID'] == 'SA0081-0018', 'PARCID'] = 'SA0081' 

			### Use PARCELNUM and ADDRGL1 to PARCID
			e911_temp = e911_temp.merge(gl_temp, left_on='PARCELNUM', right_on='PARCID', how='left')

		if town == 'WARREN':
			### Subset gl_temp to town of interest
			gl_temp = gl_temp[gl_temp['TNAME']=='Warren']

			### Remove dashes from PARCID column in gl_temp
			gl_temp['PARCID'] = gl_temp['PARCID'].map(lambda x: x.replace('-', ''))

			### Use PARCELNUM and ADDRGL1 to PARCID
			e911_temp = e911_temp.merge(gl_temp, left_on='PARCELNUM', right_on='PARCID', how='left')

					
		# If not None, then append to list
		if e911_temp is not None:
			temp_df_list.append(e911_temp)


	### Concat e911 temp dataframes back together
	e911_df = pd.concat(temp_df_list)

	### Subset columns
	e911_df = e911_df[['OBJECTID', 'TOWN', 'SITETYPE', 'REAL_FLV']]

	return e911_df


def standardize_propertyvalues(e911_df, cla_df):
	"""Standardize property values across towns using common level of appraisal.

    Parameters:
        e911_df (pandas dataframe): Joined E911 and Grand List table
        cla_df (pandas dataframe): towns' common-level of appraisal table

    Returns:
        e911_df (pandas dataframe): Joined E911 and Grand List table
    """

	print '		Standarizing property values'

	### Create column in e911_df for standarized values
	e911_df['PROPERTY_VALUE'] = np.nan

	### Iterate through towns in e911_df
	for town in e911_df['TOWN'].unique():
		### Get town's common level of appraisal
		town_cla = float(cla_df['CLA_PERCENT'][cla_df['TOWN']==town])
		
		### Fill in standardized values column
		e911_df.loc[e911_df['TOWN']==town, 'PROPERTY_VALUE'] = (
			e911_df['REAL_FLV'][e911_df['TOWN']==town] / town_cla)

	### Drop REAL_FLV column
	e911_df.drop(['REAL_FLV'])

	return e911_df

