# coding: utf-8

import os
import re
import ogr
import gdal
import scipy
import atexit
import tempfile
import shapefile
import numpy as np
import pandas as pd

from scipy import stats
from matplotlib.collections import LineCollection

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.simplefilter(action='ignore', category=RuntimeWarning)


def temporary_filename(suffix=''):
    """Get path to new temporary file that will be deleted on program exit.
        Returns a temporary filename using mkstemp. The file is deleted
        on exit using the atexit register.

    Parameters:
        suffix (string): the suffix to be appended to the temporary file

    Returns:
        path: a unique temporary filename
    """

    def remove_file(path):
        """Function to remove a file and handle exceptions to register
            in atexit."""
        try:
            os.remove(path)
        except OSError:
            # This happens if the file didn't exist, which is okay because
            # maybe we deleted it in a method
            pass

    file_handle, path = tempfile.mkstemp(suffix=suffix)
    os.close(file_handle)
    atexit.register(remove_file, path)
    
    return path


def get_raster_projection(raster_uri):
    """Get raster's projection.

    Parameters:
        raster_uri (sting): raster path URI

    Returns:
        projection (string): raster projection
    """
    
    ds = gdal.Open(raster_uri)
    projection = ds.GetProjectionRef()

    return projection 


def init_counter():
    """Initialize global variable for objective function evaluation counter.

    Parameters:
        None

    Returns:
        None
    """

    ### Initialize global counter 
    global counter
    counter = 0


def raster2array(raster_uri):
    """Convert raster to array.

    Parameters:
        raster_uri (string): path to GeoTIFF

    Returns:
        array (numpy array): a numpy array of data contained within the first 
            band of the raster_uri
        gt (list): a dataset geotransform list
    """
    
    ds = gdal.Open(raster_uri)
    gt = ds.GetGeoTransform()
    band = ds.GetRasterBand(1)
    ndval = band.GetNoDataValue()
    array = band.ReadAsArray()

    array = np.where(array == ndval, np.nan, array)
    
    return array, gt


def array2raster(array, gt, projection, raster_uri, dtype=gdal.GDT_Float32):
    """Convert array to raster.

    Parameters:
        array (numpy array): a numpy array of data contained within the first 
            band of the raster_uri
        gt (list): a dataset geotransform list
        projection (string): WKT describing the GeoTIFF projection
        raster_uri (string): path to GeoTIFF

    Returns:
        None
    """

    ### Create raster
    driver = gdal.GetDriverByName('GTiff')
    raster = driver.Create(raster_uri, array.shape[1], array.shape[0], 1, 
        dtype, options=['BIGTIFF=IF_SAFER'])
    raster.SetGeoTransform((gt[0], gt[1], 0, gt[3], 0, gt[5]))
    band = raster.GetRasterBand(1)
    
    ### Write data to raster 
    array = np.where(np.isnan(array), 9999, array)
    band.WriteArray(array)
    band.SetNoDataValue(9999)
    band.ComputeStatistics(True)
    raster.SetProjection(projection)
    
    ### Close and clean up dataset
    band.FlushCache()
    gdal.Dataset.__swig_destroy__(raster)
    band = raster = None

    return None


def get_raster_projection(raster_uri):
    ds = gdal.Open(raster_uri)
    projection = ds.GetProjectionRef()

    return projection 


def shapefile2raster(
    shapefile_uri, gt, projection, output_directory, burn_field=None):
    """Convert shapefile to raster.

    Parameters: 
        shapefile_uri (string): path to shapefile
        gt (list): a dataset geotransform list
        projection (string): WKT describing the GeoTIFF projection
        output_directory (string): a directory to export the raster
        burn_field (string): the field in the shapefile to burn values into the
            raster. If left blank, uses Z values

    Returns:
        raster_uri: path to GeoTIFF
    """
    
    # Get shapefile data
    driver = ogr.GetDriverByName('ESRI Shapefile')
    ds = driver.Open(shapefile_uri)
    lyr = ds.GetLayer()
    srs = lyr.GetSpatialRef()
    x_min, x_max, y_min, y_max = lyr.GetExtent()
    x_res = int((x_max - x_min) / gt[1])
    y_res = int((y_max - y_min) / gt[1])
    
    # Create raster
    fn = os.path.basename(shapefile_uri).split('.')[0]
    raster_uri = os.path.join(output_directory, fn + '.tif')
    raster = gdal.GetDriverByName('GTiff').Create(
        raster_uri, x_res, y_res, 1, gdal.GDT_Float32, 
        options=['BIGTIFF=IF_SAFER'])
    raster.SetGeoTransform((x_min, gt[1], 0, y_max, 0, gt[5]))
    band = raster.GetRasterBand(1)
    band.SetNoDataValue(9999)
    raster.SetProjection(projection)
    if burn_field != None:
        gdal.RasterizeLayer(
            raster, [1], lyr, options=['ATTRIBUTE=%s' % burn_field])
    else:
        gdal.RasterizeLayer(
            raster, [1], lyr, burn_values=[1])
    band.ComputeStatistics(True)

    # Close and clean up dataset
    band.FlushCache()
    gdal.Dataset.__swig_destroy__(raster)
    band = raster = None

    return raster_uri
    

def align_arrays(array, gt, aoi_array, aoi_gt, ndval=9999):
    """Align array to area-of-interest (AOI) by padding the perimeter with no 
        data values (ndval) on sides where array is too small or clipping the 
        perimeter on sides where the array is too large.

    Parameters:
        array (numpy array): array to be aligned
        gt (list): array GDAL geotransform list
        aoi_array (numpy array): AOI array to use for alignment
        aoi_gt (list): AOI array GDAL geotransform list

    Optional parameters:
        ndval (int): no data value; default set to 9999

    Returns:
        array (numpy array):
    """

    ### Get rows and cols
    rows = array.shape[0]
    cols = array.shape[1]
    aoi_rows = aoi_array.shape[0]
    aoi_cols = aoi_array.shape[1]

    ### Adjust top
    top = int((aoi_gt[3] - gt[3]) / gt[1])
    if top < 0:
        array = array[-top:, :]
    else:
        array = np.lib.pad(array, ((top, 0), (0, 0)), 'constant', 
                           constant_values=ndval)

    ### Adjust bottom
    bottom = int(aoi_rows - (rows+top))
    if bottom < 0:
        array = array[0:aoi_rows, :]
    else:
        array = np.lib.pad(array, ((0, bottom), (0, 0)), 'constant', 
                           constant_values=ndval)

    ### Adjust left
    left = int((gt[0] - aoi_gt[0]) / gt[1])
    if left < 0:
        array = array[:, -left:]
    else:
        array = np.lib.pad(array, ((0, 0), (left, 0)), 'constant', 
                           constant_values=ndval)

    ### Adjust right
    right = int(aoi_cols - (cols+left))
    if right < 0:
        array = array[:, 0:aoi_cols]
    else:
        array = np.lib.pad(array, ((0, 0), (0, right)), 'constant', 
                           constant_values=ndval)

    return array


def gini_coeff(x):
    """Calculate Gini coefficient.

        Copied from: https://stackoverflow.com/questions/
                    39512260/calculating-gini-coefficient-in-python-numpy
    
    Parameters:
        x (vector): vector of values

    Returns:
        g (float): Gini coefficient
    """

    ### Mean absolute difference
    mad = np.abs(np.subtract.outer(x, x)).mean()
   
    ### Relative mean absolute difference
    rmad = np.abs(mad / np.mean(x))
   
    ### Gini coefficient
    g = rmad / 2.0

    return g


def pareto_frontier(x, y, max_x=True, max_y=True):
    """Method to take two equally-sized lists and return just the elements which 
        lie on the Pareto frontier, sorted into order.
    
        Default behaviour is to find the maximum for both X and Y, but the 
        option is available to specify max_x=False or max_y=False to find the 
        minimum for either or both of the parameters.
        
        Copied from: oco-carbon.com/metrics/find-pareto-frontiers-in-python/

    Parameters:
        x (vector): x-axis values
        y (vector): y-axis values

    Optional parameters: 
        max_x (boolean): Maximize x, default is true
        max_y (boolean): Maximize y, default is true

    Returns:
        pfront_x (vector): x-axis values on pareto frontier
        pfront_y (vector): y-axis values on pareto frontier
    """

    ### Sort the list in either ascending or descending order of X
    myList = sorted([[x[i], y[i]] for i in range(len(x))], reverse=max_x)

    ### Start the Pareto frontier with the first value in the sorted list
    p_front = [myList[0]]    
    
    ### Loop through the sorted list
    for pair in myList[1:]:
        if max_y: 
            if pair[1] > p_front[-1][1]: # Look for higher values of Y…
                p_front.append(pair) # … and add them to the Pareto frontier
        
        else:
            if pair[1] < p_front[-1][1]: # Look for lower values of Y…
                p_front.append(pair) # … and add them to the Pareto frontier
   
    ### Turn resulting pairs back into a list of x and y
    pfront_x = [pair[0] for pair in p_front]
    pfront_y = [pair[1] for pair in p_front]

    return pfront_x, pfront_y


def pareto_frontier_indexed(costs, return_mask=True):
    """
    
    Copied from: https://stackoverflow.com/questions/32791911/...
                 fast-calculation-of-pareto-front-in-python

    Parameters:
        costs: An (n_points, n_costs) array
        return_mask: True to return a mask, False to return integer indices of 
                     efficient points.
    
    Returns: 
        An array of indices of pareto-efficient points.
        If return_mask is True, this will be an (n_points, ) boolean array
        Otherwise it will be a (n_efficient_points, ) integer array of indices.
    """

    is_efficient = np.arange(costs.shape[0])
    n_points = costs.shape[0]
    next_point_index = 0  

    while next_point_index<len(costs):
        nondominated_point_mask = np.any(costs<=costs[next_point_index], axis=1)
        is_efficient = is_efficient[nondominated_point_mask]  
        costs = costs[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index])+1

    if return_mask:
        is_efficient_mask = np.zeros(n_points, dtype = bool)
        is_efficient_mask[is_efficient] = True
        return is_efficient_mask
    
    else:
        return is_efficient


def obj2json(obj):
    """Dump object to json.

    Parameters:
        obj (object): some object

    Returns:
        obj (object): some object in json format or dictionary
    """
    try:
        return obj.toJSON()
    except:
        return obj.__dict__


def test_objfunc(indv, args_dict):
    """Test objective function.

        Randomly initialize list of integers in range [0,N_interventions] with 
        length of scenario. Compare scenario to randomly initialized lists. 
        Where values are equal, add one to fitness1, subtract one from fitness2.

    Parameters:
        indv (string): encoding for individual scenario
        args_dict (dict): dictionary containing objective function parameters

    Returns:
        fitnesses_list (list): list objectives' fitness values
    """

    ### Initialize
    N_objectives = len(args_dict['obj_metrics'])
    N_interventions = args_dict['N_interventions']
    fitnesses_list = []    
   
    ### Loop through objectives
    for i in range(N_objectives):
        ### Randomly initialize globally optimal scenario
        np.random.seed(seed=i)
        global_optimum = np.random.randint(N_interventions + 1, size=len(indv))

        ### Compare indv to globally optimal indv and calculate fitness
        ### for indv
        fitness = float(np.sum(np.where(global_optimum==np.array(indv), 1, 0)))
        fitness = (fitness / len(indv)) * 100

        ### Append fitness to fitnesses_list
        fitnesses_list.append(fitness)

    return fitnesses_list


class shp():
    def __init__(self, shp_uri):
        self.shp = shapefile.Reader(shp_uri)
        self.bbox = self.shp.bbox
        self.records = self.shp.records()
        self.shapes = self.shp.shapes()
        self.fields = [field[0] for field in self.shp.fields[1:]]
        self.records = self.shp.records()
        self.df = None

    def parse_shp(self):    
        df = pd.DataFrame(self.records, columns=self.fields)
        paths = pd.Series()
        for i, shape in enumerate(self.shapes):
            shape_df = np.array(shape.points)
            if len(shape.parts) == 1:
                segs = [shape_df, ]
            else:
                segs = []
                for y in range(1, len(shape.parts)):
                    index = shape.parts[y-1]
                    index2 = shape.parts[y]
                    segs.append(shape_df[index:index2])
                segs.append(shape_df[index2:])
            paths.loc[len(paths)] = LineCollection(segs)
        df['paths'] = paths
    
        return df


def xs_nearestneighbor():

    ### Initialize input data
    stem = '/Users/jessegourevitch/Documents/UVM/Dissertation/Ch2_HECRAS/Data/'
    xs_raster_uri = stem + 'Cross_Sections/lewiscreek_xs_10m.tif'
    xs_polygon_raster_uri = stem + 'Cross_Sections/lewiscreek_xs_polygon_10m.tif'

    ### Read rasters to arrays
    xs_fid_array, xs_fid_gt = raster2array(xs_raster_uri)
    xs_polygon_array, xs_polygon_gt = raster2array(xs_polygon_raster_uri)

    ### Align arrays
    xs_fid_array = align_arrays(xs_fid_array, xs_fid_gt, xs_polygon_array, 
                                xs_polygon_gt, ndval=np.nan)

    ### Do stuff with arrays
    X = np.indices(xs_fid_array.shape)[1]
    Y = np.indices(xs_fid_array.shape)[0]
    min_dist = np.full(xs_fid_array.shape, float(10**6))
    min_fid  = np.full(xs_fid_array.shape, -1.0)

    for fid in np.unique(xs_fid_array[~np.isnan(xs_fid_array)]):
        print fid
        x_list = list(np.nonzero(xs_fid_array==fid)[1])
        y_list = list(np.nonzero(xs_fid_array==fid)[0])    

        for x, y in zip(x_list, y_list):
            dist = np.sqrt((X - x)**2 + (Y - y)**2)
            min_fid[dist < min_dist] = fid
            min_dist[dist < min_dist] = dist[dist < min_dist]

    min_fid = np.where(xs_polygon_array==0, min_fid, np.nan)
    proj = get_raster_projection(xs_polygon_raster_uri)
    output_uri = stem + 'Cross_Sections/lewiscreek_xs_polygon_nearestneighbor_10m.tif'
    array2raster(min_fid, xs_polygon_gt, proj, output_uri, dtype=gdal.GDT_Int16)

    return None


def matrix_entropy(gen, matrix, entropy_dict={}):
    """Get the row and column entropy for generation matrix.

    Parameters:
        gen (int): generation number
        matrix (matrix): matrix of best scenarios

    Optional parameters:
        entropy_dict (dict): 

    Returns: 
        None
    """
    
    ### Gen column entropy
    col_entropy = []
    for col in range(matrix.shape[0]):
        count_data = np.unique(matrix[col,:].tolist(), return_counts=True)
        p_dat = count_data[1] / sum(count_data[1])
        entropy = stats.entropy(p_dat, base=2)
        col_entropy.append(entropy)

    ### Gen row entropy
    row_entropy = []
    for row in range(matrix.shape[1]):
        count_data = np.unique(matrix[:,row].tolist(), return_counts=True)
        entropy = stats.entropy(count_data[1] / sum(count_data[1]), base=2)
        row_entropy.append(entropy)

    entropy_dict.update({gen:{'col_entropy':col_entropy,
                              'row_entropy':row_entropy}})

    return None


def genpareto_entropy(gen_list, df):
    """ Get the entropy dict from a dataframe for the generations in gen_list. 
        Also return the entropy for the pareto optimal solutions. 

    Parameters:
        gen_list (list): list of generations
        df (Pandas dataframe): summary logbook dataframe

    Returns:
        entropy_dict (dict): dictionary of entropy values for generations
        best_ind_ent_dict (dict): dictionary of 
    """
    
    regex = re.compile(r'\d{1}')
    gen_dict = {}
    best_ind_ent_dict = {}
    entropy_dict = {}
    
    for g in gen_list:
        gen_list = []
        temp_df = df[df.GENERATION==g]
        for row in temp_df.index:
            gen_list.append(regex.findall(temp_df.at[row,'INDIVIDUAL']))
        
        matrix_entropy(g, np.matrix(gen_list), entropy_dict)
        
        ### Get the pareto optimal individuals
        p_return = pareto_frontier(temp_df[['FITNESS0','FITNESS1']].values[:,0], 
            temp_df[['FITNESS0','FITNESS1']].values[:,1], max_x=False)
    
        X1s = [fit for fit in p_return[1]]
        X0s = [fit for fit in p_return[0]]
        
        ### Get the pareto optimal individuals 
        pareto_dict = {}
        fitness_tuple = list(zip(X1s,X0s))
        for fit_pair in fitness_tuple:
            pareto_dict.update({fit_pair:temp_df[(temp_df.FITNESS0==fit_pair[1])
                & (temp_df.FITNESS1==fit_pair[0])].INDIVIDUAL.values})
        
        ### Get the genome for each individual
        best_scenarios = []
        for opt_ind in list(pareto_dict.values()):
            best_scenarios.append(regex.findall(opt_ind[0]))
            
        matrix_entropy(g, np.matrix(best_scenarios), best_ind_ent_dict)
        
    return (entropy_dict, best_ind_ent_dict)


def nondom_inds(df, gen_list):
    """ Get non-dominated individuals from each generation. 
    
    """
    
    Xzero = []
    Xone = [] 
    
    for g in gen_list:
        temp_df = df[df['GENERATION']==g]
        
        pfront_x, pfront_y = pareto_frontier(
            temp_df[['COST','AVOIDEDCOST_EW0']].values[:,0], 
            temp_df[['COST','AVOIDEDCOST_EW0']].values[:,1], max_x=False)
    
        X0s = [fit for fit in pfront_x]
        X1s = [fit for fit in pfront_y]

        Xzero.append(X0s)
        Xone.append(X1s)
    
    return (Xzero, Xone)


        